# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define the register of univiariate colormaps."""

import os
import tempfile

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from pycolorbar.settings.colormap_io import read_cmap_dict, write_cmap_dict
from pycolorbar.settings.utils import get_auxiliary_categories
from pycolorbar.utils.yaml import list_yaml_files

# Matplotlib Registry Initialization:
# - In the __init__ : from matplotlib.cm import _colormaps as colormaps
# - Instead of singleton pattern, uses globals().updates() for module-level instance
#   --> https://github.com/matplotlib/matplotlib/blob/v3.8.2/lib/matplotlib/cm.py#L59
#   --> https://github.com/matplotlib/matplotlib/blob/v3.8.2/lib/matplotlib/cm.py#L236
#
#   _colormaps = ColormapRegistry()
#   globals().update(_colormaps)


class ColormapRegistry:
    """
    A singleton class to manage colormap registrations.

    This class provides methods to register colormaps, remove them, retrieve colormap file paths,
    and add new colormaps to a temporary directory for further processing.

    Attributes
    ----------
    _instance : ColormapRegistry
        The singleton instance of the ColormapRegistry.
    registry : dict
        The dictionary holding the registered colormap names and their corresponding configuration YAML file paths.
    tmp_dir : str
        The path of a temporary directory where colormap YAML files are stored when specifying a colormap
        on-the-fly with add_cmap_dict(cmap_dict).
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            #  cls._instance = super(ColormapRegistry, cls).__new__(cls)
            cls._instance.registry = {}
            # Create temporary path
            cls._instance.tmp_dir = None
        return cls._instance

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls()  # this will call __new__
        return cls._instance

    def reset(self):
        """Clears the entire colormap registry."""
        self.registry.clear()

    @property
    def names(self):
        """List the names of all registered colormaps."""
        return sorted(list(self.registry))

    def __contains__(self, item):
        """Test registration of colormap in the registry."""
        return item in self.names

    def _check_if_cmap_in_use(self, name, force, verbose):
        if name in self.registry:
            if force and verbose:
                print(f"Warning: Overwriting existing colormap '{name}'")
            if not force:
                raise ValueError(f"A colormap named '{name}' already exists. To allow overwriting, set 'force=True'.")

    def register(self, filepath: str, verbose: bool = True, force: bool = True):
        """
        Register a colormap by its name and file path.

        Parameters
        ----------
        filepath : str
            The file path where the colormap's YAML file is located.
            The name of the colormap correspond to the name of the YAML file !
        verbose : bool, optional
            If True, the method will print a warning when overwriting an existing colormap. The default is True.
        force : bool, optional
            If True, it allow to overwrites an existing colormap. The default is True.
            If False, it raise an error if attempting to overwrite an existing colormap.

        Notes
        -----
        If a colormap with the same name already exists, it will be overwritten, and a warning will be printed.
        The validity of the colormap's YAML file is not validated !
        """
        # Check file exists
        if not os.path.isfile(filepath):
            raise ValueError(f"The colormap configuration YAML file {filepath} does not exist.")
        # Define colormap name
        name = os.path.splitext(os.path.basename(filepath))[0]
        # Check if the name is already used
        self._check_if_cmap_in_use(name=name, force=force, verbose=verbose)
        # Register
        self.registry[name] = filepath

    def add_cmap_dict(self, cmap_dict: dict, name: str, verbose: bool = True, force=True):
        """
        Add a colormap to the registry by providing a colormap dictionary and the colormap name.
        A temporary file YAML configuration file is created in self.tmp_dir.

        Parameters
        ----------
        cmap_dict : dict
            The colormap dictionary containing the colormap's configuration.
        name : str
            The name of the colormap.
        verbose : bool, optional
            If True, the method will print a warning when overwriting an existing colormap. The default is True.
        force : bool, optional
            If True, it allow to overwrites an existing colormap. The default is True.
            If False, it raise an error if attempting to overwrite an existing colormap.

        Notes
        -----
        If a colormap with the same name already exists, it will be overwritten, and a warning will be printed.
        The YAML file for the colormap is stored in a temporary directory.
        """
        # Create a temporary directory if not yet initiated
        if self.tmp_dir is None:
            self.tmp_dir = tempfile.mkdtemp(prefix="pycolorbar_cmaps_")
        # Check if the name is already used
        self._check_if_cmap_in_use(name=name, force=force, verbose=verbose)
        # Define filepath
        filename = f"{name}.yaml"
        filepath = os.path.join(self.tmp_dir, filename)
        # Write cmap_dict (and validate)
        write_cmap_dict(cmap_dict, filepath=filepath, force=True)
        # Update registry
        self.registry[name] = filepath

    def unregister(self, name: str):
        """
        Remove a colormap from the registry.

        Parameters
        ----------
        name : str
            The name of the colormap to remove.

        Raises
        ------
        ValueError
            If the colormap with the specified name is not registered.
        """
        if name in self.registry:
            _ = self.registry.pop(name)
        else:
            raise ValueError(f"The colormap {name} is not registered in pycolorbar.")

    def get_cmap_filepath(self, name: str):
        """
        Retrieve the colormap's YAML configuration file path.

        Parameters
        ----------
        name : str
            The name of the colormap.

        Returns
        -------
        str
            The colormap's YAML configuration file path.

        Raises
        ------
        ValueError
            If the colormap with the specified name is not registered.
        """
        # Remove _r suffix
        if name.endswith("_r"):
            name = name[:-2]
        # Retrieve filepath
        if name not in self.registry:
            raise ValueError(f"The {name} colormap is not registered in pycolorbar.")
        return self.registry[name]

    def get_cmap_dict(self, name: str):
        """
        Retrieve the validated colormap dictionary of a registered colormap.

        Parameters
        ----------
        name : str
            The name of the colormap.

        Returns
        -------
        dict
            The validated colormap dictionary.

        Raises
        ------
        ValueError
            If the colormap configuration is invalid or cannot be read.
        """
        filepath = self.get_cmap_filepath(name)
        return read_cmap_dict(filepath)

    def get_cmap(self, name: str):
        """
        Retrieve the colormap.

        Parameters
        ----------
        name : str
            The name of the colormap.

        Returns
        -------
        matplotlib.Colormap
            The matplotlib colormap.

        Raises
        ------
        ValueError
            If the colormap configuration is invalid or cannot be read.
        """
        from pycolorbar.settings.colormap_utility import create_cmap

        cmap_dict = self.get_cmap_dict(name)
        cmap = create_cmap(name=name, cmap_dict=cmap_dict)
        if name.endswith("_r"):
            cmap = cmap.reversed(name)
        return cmap

    def validate(self, name: str = None):
        """
        Validate the registered colormaps. If a specific name is provided, only that colormap is validated.

        Parameters
        ----------
        name : str, optional
            The name of a specific colormap to validate. If None, all registered colormaps are validated.

        Raises
        ------
        ValueError
            If any of the validated colormaps have invalid configurations.

        Notes
        -----
        Invalid colormap configurations are reported.
        """
        if isinstance(name, str):
            if name not in self.names:
                raise ValueError(f"{name} is not a registered colormap.")
            names = [name]

        else:
            names = self.names

        # Validate colormaps
        wrong_names = []
        for name in names:
            try:
                _ = self.get_cmap_dict(name)
            except Exception as e:
                wrong_names.append(name)
                print(f"{name} has an invalid configuration: {e}")
                print("")
        if wrong_names:
            raise ValueError(f"The {wrong_names} colormaps have invalid configurations.")
        return

    def to_yaml(self, name, filepath, force=False):
        """Write the colormap configuration to a YAML file."""
        cmap_dict = self.get_cmap_dict(name)
        write_cmap_dict(cmap_dict=cmap_dict, filepath=filepath, force=force)

    def _get_subset_names(self, category):
        names = []
        for name in self.names:
            cmap_dict = self.get_cmap_dict(name)
            categories = get_auxiliary_categories(cmap_dict)
            categories = [cat.upper() for cat in categories]
            if category.upper() in categories:
                names.append(name)
        return names

    def available(self, category=None, include_reversed=False):
        """List the name of available colormaps for a specific category."""
        if category is None:
            names = self.names
        else:
            names = self._get_subset_names(category=category)

        if include_reversed:
            names = [name + "_r" for name in names] + names
        return sorted(names)

    def show_colormap(self, name):
        """Display a colormap."""
        from pycolorbar.settings.colormap_visualization import plot_colormap

        cmap = self.get_cmap(name)
        plot_colormap(cmap)

    def show_colormaps(self, category=None, include_reversed=False, subplot_size=None):
        """Display available colormaps (optionally of a specific category)"""
        from pycolorbar.settings.colormap_visualization import plot_colormaps

        # Retrieve available colormaps (of a given category)
        names = self.available(category=category, include_reversed=include_reversed)
        if len(names) == 0:
            raise ValueError("No colormaps are yet registered in the pycolorbar ColormapRegistry.")

        # If only 1 colormap registered, plot it with the other method
        if len(names) == 1:
            self.show_colormap(name=names[0])
            return None

        # Else, retrieve colormaps to display
        cmaps = [self.get_cmap(name) for name in sorted(names)]

        # Display colormaps
        plot_colormaps(cmaps, subplot_size=subplot_size)


def register_colormaps(directory: str, name: str = None, verbose: bool = True, force: bool = True):
    """
    Register all colormap YAML files present in the specified directory (if name=None).

    This function assumes that all YAML files present in the directory are
    valid pycolorbar colormaps configuration files.

    Parameters
    ----------
    directory : str
        The directory where colormap YAML files are located.
    name : str, optional
        The specific name of a colormap to register. If None, all colormaps in the directory are registered.
    force : bool, optional
        If True, it allow to overwrites existing colormaps. The default is True.
        If False, it raise an error if attempting to overwrite an existing colormap.
    verbose : bool, optional
        If True, the method will print a warning when overwriting existing colormaps. The default is True.
    """
    colormaps = ColormapRegistry.get_instance()

    # List the colormap YAML files to register
    if name is not None:
        filepaths = [os.path.join(directory, f"{name}.yaml")]
    else:
        # List all YAML files in the directory
        filepaths = list_yaml_files(directory)

    # Add colormaps to the ColormapRegistry
    for filepath in filepaths:
        colormaps.register(filepath, verbose=verbose, force=force)


def register_colormap(filepath: str, verbose: bool = True, force: bool = True):
    """
    Register a single colormap YAML file.

    Parameters
    ----------
    filepath : str
        The file path where the colormap's YAML file is located.
        The name of the colormap correspond to the name of the YAML file !
    force : bool, optional
        If True, it allow to overwrites an existing colormap. The default is True.
        If False, it raise an error if attempting to overwrite an existing colormap.
    verbose : bool, optional
        If True, the method will print a warning when overwriting an existing colormap. The default is True.

    Raises
    ------
    ValueError
        If the specified colormap YAML file is not available in the directory or
        if trying to register an already registered colormap and `force=False`.
    """
    colormaps = ColormapRegistry.get_instance()
    colormaps.register(filepath, verbose=verbose, force=force)


def get_cmap_dict(name):
    """
    Retrieve the validated colormap dictionary of a registered colormap.

    Parameters
    ----------
    name : str
        The name of the colormap.

    Returns
    -------
    dict
        The validated colormap dictionary.

    Raises
    ------
    Exception
        If the colormap configuration is invalid or cannot be read.
    """
    colormaps = ColormapRegistry.get_instance()
    return colormaps.get_cmap_dict(name)


def get_cmap(name: str = None, lut: int = None):
    """
    Get a colormap instance.

    Parameters
    ----------
    name : str
        The name of a colormap known to pycolorbar or Matplotlib.
        If the name ends with the suffix `_r`, the colormap is reversed.
    lut : int or None
        If not `None` (the default), the colormap will be resampled to have *lut* entries in the lookup table.
        If the name ends with the suffix `_r`, the resampling is done after reversing the colormap.

    Returns
    -------
    Colormap
    """
    colormaps = ColormapRegistry.get_instance()

    # Use default matplotlib colormap
    if name is None:
        return plt.get_cmap(name=name, lut=lut)

    # If already colormap provided, return it
    if isinstance(name, mpl.colors.Colormap):
        return name

    # Retrieve registered colormaps names
    pycolorbar_registered_names = colormaps.names + [s + "_r" for s in colormaps.names]
    mpl_registered_names = plt.colormaps()

    # Check if reversed colormap
    has_r_suffix = name.endswith("_r")
    # Pycolorbar colormap
    if name in pycolorbar_registered_names:
        cmap = colormaps.get_cmap(name)
        if has_r_suffix:
            cmap = cmap.reversed()
        if lut is not None:
            cmap = cmap.resampled(lut)
        return cmap

    # Matplotlib registered colormap
    if name in mpl_registered_names:
        return plt.get_cmap(name=name, lut=lut)
    # Unavailable colormap
    else:
        raise ValueError(
            f"{name} is not registered in pycolorbar and matplotlib !\n "
            f"Valid matplotlib colormap are {mpl_registered_names}.\n "
            f"Valid pycolorbar colormap are {pycolorbar_registered_names}."
        )


def _get_mpl_cmap_by_category(category):
    """Return matplotlib colormap names by category.

    See: https://matplotlib.org/stable/users/explain/colors/colormaps.html#choosing-colormaps
    """
    type_dict = {
        "PERCEPTUAL": ["viridis", "plasma", "inferno", "magma", "cividis"],  # PERCEPTUALLY UNIFORM
        "SEQUENTIAL": [
            "viridis",
            "plasma",
            "inferno",
            "magma",
            "cividis",
            "Greys",
            "Purples",
            "Blues",
            "Greens",
            "Oranges",
            "Reds",
            "YlOrBr",
            "YlOrRd",
            "OrRd",
            "PuRd",
            "RdPu",
            "BuPu",
            "GnBu",
            "PuBu",
            "YlGnBu",
            "PuBuGn",
            "BuGn",
            "YlGn",
            "binary",
            "gist_yarg",
            "gist_gray",
            "gray",
            "bone",
            "pink",
            "spring",
            "summer",
            "autumn",
            "winter",
            "cool",
            "Wistia",
            "hot",
            "afmhot",
            "gist_heat",
            "copper",
        ],
        "DIVERGING": [
            "PiYG",
            "PRGn",
            "BrBG",
            "PuOr",
            "RdGy",
            "RdBu",
            "RdYlBu",
            "RdYlGn",
            "Spectral",
            "coolwarm",
            "bwr",
            "seismic",
        ],
        "QUALITATIVE": [
            "Pastel1",
            "Pastel2",
            "Paired",
            "Accent",
            "Dark2",
            "Set1",
            "Set2",
            "Set3",
            "tab10",
            "tab20",
            "tab20b",
            "tab20c",
        ],
        "CYCLIC": ["twilight", "twilight_shifted", "hsv"],
    }
    type_dict["CATEGORICAL"] = type_dict["QUALITATIVE"]
    if category.upper() in type_dict:
        names = type_dict[category.upper()]
    else:
        names = []
    return names


def _get_matplotlib_cmaps(category=None, include_reversed=False):
    if category is None:
        names = plt.colormaps()
        names = [name for name in names if not name.endswith("_r")]
    else:
        names = _get_mpl_cmap_by_category(category=category)
    # Include names with _r suffix
    if include_reversed:
        names = [name + "_r" for name in names] + names
    return names


def available_colormaps(category=None, include_reversed=False):
    colormaps = ColormapRegistry.get_instance()
    names = colormaps.available(category=category, include_reversed=include_reversed)
    names += _get_matplotlib_cmaps(category=category, include_reversed=include_reversed)
    names = sorted(np.unique(names))
    return names



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define the register of univiariate colorbars."""

import os

from pycolorbar.settings.colorbar_io import read_cbar_dicts, write_cbar_dicts
from pycolorbar.settings.colorbar_validator import validate_cbar_dict
from pycolorbar.settings.utils import get_auxiliary_categories
from pycolorbar.utils.yaml import list_yaml_files


class ColorbarRegistry:
    """
    A singleton class to manage colorbar registrations.

    This class provides methods to register colorbars settings, add new one on-the-fly,
    and to remove them.

    Attributes
    ----------
    _instance : ColorbarRegistry
        The singleton instance of the ColorbarRegistry.
    registry : dict
        The dictionary holding the registered colorbar settings.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            #  cls._instance = super(ColorbarRegistry, cls).__new__(cls)
            cls._instance.registry = {}
        return cls._instance

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls()  # this will call __new__
        return cls._instance

    def reset(self):
        """Clears the entire Colorbar registry."""
        self.registry.clear()

    @property
    def names(self):
        """List the names of all registered colorbars settings."""
        return sorted(list(self.registry))

    def __contains__(self, item):
        """Test registration of a colorbar in the registry."""
        return item in self.names

    def _check_if_cbar_in_use(self, name, force, verbose):
        if name in self.registry:
            if force and verbose:
                print(f"Warning: Overwriting existing colorbar '{name}'")
            if not force:
                raise ValueError(
                    f"A colorbar setting named '{name}' already exists. To allow overwriting, set 'force=True'."
                )

    def register(self, filepath: str, verbose: bool = True, force: bool = True, validate=False):
        """
        Register colorbar(s) configuration(s) defined in a YAML file.

        Parameters
        ----------
        filepath : str
            The YAML file path where the colorbar(s) settings are specified.
            A single YAML file can contain the configuration of multiple colorbars !
            The name of the YAML files it's not used !
        force : bool, optional
            If True, it allow to overwrites existing colorbar settings. The default is True.
            If False, it raise an error if attempting to overwrite an existing colorbar.
        verbose : bool, optional
            If True, the method will print a warning when overwriting existing colorbars. The default is True.
        validate: bool, optional
            Whether to validate the colorbar configuration file before registering.
            The default is False.
        Notes
        -----
        If a a colorbar configuration with the same name already exists, it will be overwritten.
        The validity of the colorbar(s) configuration(s) is not validated at registration !
        Use `pycolorbar.colorbars.validate()` to validate the registered colorbars.
        """
        # Check file exists
        if not os.path.isfile(filepath):
            raise ValueError(f"The colorbars configuration YAML file {filepath} does not exist.")
        # Read colorbars settings
        cbar_dicts = read_cbar_dicts(filepath=filepath)
        # Register colorbars settings
        for name, cbar_dict in cbar_dicts.items():
            if validate:
                cbar_dict = validate_cbar_dict(cbar_dict=cbar_dict, name=name)
            self._check_if_cbar_in_use(name=name, force=force, verbose=verbose)
            self.registry[name] = cbar_dict

    def add_cbar_dict(self, cbar_dict: dict, name: str, verbose: bool = True, force: bool = True):
        """
        Add a colorbar configuration to the registry by providing a colorbar dictionary.

        Parameters
        ----------
        cbar_dict : dict
            The colorbar dictionary containing the colorbar's configuration.
        name : str
            The name of the colorbar.
        verbose : bool, optional
            If True, the method will print a warning when overwriting an existing colorbar. The default is True.
        force : bool, optional
            If True, it allow to overwrites existing colorbar settings. The default is True.
            If False, it raise an error if attempting to overwrite an existing colorbar.

        Notes
        -----
        If a colorbar with the same name already exists, it will be overwritten.
        The configuration is validated when adding a colorbar configuration with this method !
        """
        # Check if the name is already used
        self._check_if_cbar_in_use(name=name, force=force, verbose=verbose)
        # Validate cbar_dict
        cbar_dict = validate_cbar_dict(cbar_dict, name=name)
        # Update registry
        self.registry[name] = cbar_dict

    def unregister(self, name: str):
        """
        Remove a specific colorbar configuration from the registry.

        Parameters
        ----------
        name : str
            The name of the colorbar's configuration to remove.

        Raises
        ------
        ValueError
            If the colorbar with the specified name is not registered.
        """
        if name in self.registry:
            _ = self.registry.pop(name)
        else:
            raise ValueError(f"The colorbar configuration for {name} is not registered in pycolorbar.")

    def get_cbar_dict(self, name: str, resolve_reference=True, validate=True):
        """
        Retrieve the colorbar dictionary of a registered colorbar.

        Parameters
        ----------
        name : str
            The name of the colorbar.
        resolve_reference: bool
            Determines the behavior when the colorbar dictionary contains the 'reference' keyword.
            If True, the function resolves the reference by returning the actual colorbar dictionary
            that the reference points to.
            If False, the function returns the original colorbar dictionary, including the 'reference' keyword.
            The default is True.
        validate: bool
            Whether to validate the colorbar dictionary.
            The default is True.

        Returns
        -------
        dict
            The validated colorbar dictionary.

        Raises
        ------
        ValueError
            If the colorbar configuration is not registered.
        """
        if name not in self.registry:
            raise ValueError(f"The colorbar configuration for {name} is not registered in pycolorbar.")
        cbar_dict = self.registry[name]
        if validate:
            cbar_dict = validate_cbar_dict(cbar_dict, name=name, resolve_reference=resolve_reference)
        return cbar_dict

    def get_cmap(self, name):
        """
        Retrieve the colormap of a registered colorbar.

        Parameters
        ----------
        name : str
            The name of the colorbar.

        Returns
        -------
        matplotlib.colors.Colormap
            The matplotlib Colormap.

        Notes
        -----
        This function also sets the over/under and bad colors specified in the colorbar configuration.

        """
        from pycolorbar.settings.colorbar_utility import get_cmap

        cbar_dict = self.get_cbar_dict(name=name, resolve_reference=True)
        return get_cmap(cbar_dict)

    def to_yaml(self, filepath, names=None, force=False, sort_keys=False):
        write_cbar_dicts(
            cbar_dicts=self.registry,
            filepath=filepath,
            names=names,
            force=force,
            sort_keys=sort_keys,
        )

    def validate(self, name: str = None):
        """
        Validate the registered colorbars. If a specific name is provided, only that colorbar is validated.

        Parameters
        ----------
        name : str, optional
            The name of a specific colorbar to validate. If None, all registered colorbars are validated.

        Raises
        ------
        ValueError
            If any of the validated colorbars have invalid configurations.

        Notes
        -----
        Invalid colorbar configurations are reported.
        """
        # TODO: allow for list of names ?
        if isinstance(name, str):
            names = [name]
        else:
            names = self.names

        # Validate colorbars
        wrong_names = []
        for name in names:
            try:
                cbar_dict = self.get_cbar_dict(name, resolve_reference=False)
                _ = validate_cbar_dict(cbar_dict=cbar_dict, name=name)
            except Exception as e:
                wrong_names.append(name)
                print(f"{name} has an invalid configuration: {e}")
                print("")
        if wrong_names:
            raise ValueError(f"The {wrong_names} colorbars have invalid configurations.")
        return

    def get_standalone_settings(self):
        """Return the colorbar settings names which are not a reference to another colorbar."""
        names = []
        for name in self.names:
            cbar_dict = self.get_cbar_dict(name, resolve_reference=False)
            if "reference" not in cbar_dict:
                names.append(name)
        return names

    def get_referenced_settings(self):
        """Return the colorbar settings names which a reference to another colorbar."""
        names = []
        for name in self.names:
            cbar_dict = self.get_cbar_dict(name, resolve_reference=False)
            if "reference" in cbar_dict:
                names.append(name)
        return names

    def available(self, category=None, exclude_referenced=False):
        """List the name of available colorbars for a specific category."""
        if exclude_referenced:
            names = self.get_standalone_settings()
        else:
            names = self.names

        if category is None:
            return names

        # Subset names by category
        cat_names = []
        for name in names:
            cbar_dict = self.get_cbar_dict(name, resolve_reference=True)
            categories = get_auxiliary_categories(cbar_dict)
            categories = [cat.upper() for cat in categories]
            if category.upper() in categories:
                cat_names.append(name)
        return cat_names

    def show_colorbar(self, name, user_plot_kwargs={}, user_cbar_kwargs={}, fig_size=(6, 1)):
        """Display a colorbar (updated with optional user arguments)."""
        from pycolorbar.settings.colorbar_visualization import plot_colorbar

        plot_kwargs, cbar_kwargs = self.get_plot_kwargs(
            name=name, user_plot_kwargs=user_plot_kwargs, user_cbar_kwargs=user_cbar_kwargs
        )
        plot_colorbar(plot_kwargs=plot_kwargs, cbar_kwargs=cbar_kwargs, ax=None, subplot_size=fig_size)

    def show_colorbars(self, category=None, exclude_referenced=True, subplot_size=None):
        """Display available colorbars (optionally of a specific category)"""
        from pycolorbar.settings.colorbar_visualization import plot_colorbars

        # TODO: allow for names subset ?

        # Retrieve available (of a given category) colorbars settings
        names = self.available(category=category, exclude_referenced=exclude_referenced)
        if len(names) == 0:
            raise ValueError("No colorbars are yet registered in the pycolorbar ColorbarRegistry.")

        # If only 1 colormap registered, plot it with the other method
        if len(names) == 1:
            self.show_colorbar(name=names[0])
            return None

        # Display colorbars
        list_args = [[name] + list(self.get_plot_kwargs(name=name)) for name in names]
        plot_colorbars(list_args, subplot_size=subplot_size)

    def get_plot_kwargs(self, name=None, user_plot_kwargs={}, user_cbar_kwargs={}):
        """Get pycolorbar plot kwargs (updated with optional user arguments)."""
        from pycolorbar.settings.colorbar_utility import (
            get_plot_cbar_kwargs,
            update_plot_cbar_kwargs,
        )

        if not isinstance(name, (str, type(None))):
            raise TypeError("Expecting the colorbar setting name.")

        try:
            cbar_dict = self.get_cbar_dict(name)
        except Exception:
            cbar_dict = {}

        # Retrieve defaults pycolorbar kwargs
        plot_kwargs, cbar_kwargs = get_plot_cbar_kwargs(cbar_dict)
        plot_kwargs, cbar_kwargs = update_plot_cbar_kwargs(
            default_plot_kwargs=plot_kwargs,
            default_cbar_kwargs=cbar_kwargs,
            user_plot_kwargs=user_plot_kwargs,
            user_cbar_kwargs=user_cbar_kwargs,
        )
        return plot_kwargs, cbar_kwargs


def register_colorbars(directory: str, verbose: bool = True, force: bool = True):
    """
    Register all colorbar YAML files present in the specified directory (if name=None).

    This function assumes that all YAML files present in the directory are
    valid pycolorbar colorbars configuration files.

    Parameters
    ----------
    directory : str
        The directory where colorbar YAML files are located.
    force : bool, optional
        If True, it allow to overwrites existing colorbar settings. The default is True.
        If False, it raise an error if attempting to overwrite an existing colorbar.
    verbose : bool, optional
        If True, the method will print a warning when overwriting existing colorbars. The default is True.

    Notes
    -----
    If a a colorbar configuration with the same name already exists and `force=True`, it will be overwritten.
    The validity of the colorbar(s) configuration(s) is not validated at registration !
    Use `pycolorbar.colorbars.validate()` to validate the registered colorbars.
    """
    # List the colorbar YAML files to register
    filepaths = list_yaml_files(directory)

    # Add colorbars to the ColorbarRegistry
    colorbars = ColorbarRegistry.get_instance()
    for filepath in filepaths:
        colorbars.register(filepath, force=force, verbose=verbose)


def register_colorbar(filepath: str, verbose: bool = True, force: bool = True):
    """
    Register a single colorbar YAML file.

    Parameters
    ----------
    filepath : str
        The file path where the colorbar's YAML file is located.
    force : bool, optional
        If True, it allow to overwrites existing colorbar settings. The default is True.
        If False, it raise an error if attempting to overwrite an existing colorbar.
    verbose : bool, optional
        If True, the method will print a warning when overwriting existing colorbars. The default is True.

    Raises
    ------
    ValueError
        If the specified colorbar YAML file is not available in the directory.

    Notes
    -----
    If a a colorbar configuration with the same name already exists and `force=True`, it will be overwritten.
    The validity of the colorbar(s) configuration(s) is not validated at registration !
    Use `pycolorbar.colorbars.validate()` to validate the registered colorbars.
    """
    colorbars = ColorbarRegistry.get_instance()
    colorbars.register(filepath, verbose=verbose, force=force)


def get_cbar_dict(name, resolve_reference=True):
    """
    Retrieve the validated colorbar dictionary of a registered colorbar.

    Parameters
    ----------
    name : str
        The name of the colorbar.
    resolve_reference: bool
        Determines the behavior when the colorbar dictionary contains the 'reference' keyword.
        If True, the function resolves the reference by returning the
        actual colorbar dictionary that the reference points to.
        If False, the function returns the original colorbar dictionary, including the 'reference' keyword.
        The default is True.

    Returns
    -------
    dict
        The validated colorbar dictionary.

    """
    colorbars = ColorbarRegistry.get_instance()
    return colorbars.get_cbar_dict(name, resolve_reference=resolve_reference)


def get_plot_kwargs(name=None, user_plot_kwargs={}, user_cbar_kwargs={}):
    colorbars = ColorbarRegistry.get_instance()
    return colorbars.get_plot_kwargs(name=name, user_plot_kwargs=user_plot_kwargs, user_cbar_kwargs=user_cbar_kwargs)


def available_colorbars(category=None, exclude_referenced=False):
    colorbars = ColorbarRegistry.get_instance()
    names = colorbars.available(category=category, exclude_referenced=exclude_referenced)
    return names



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to visualize univariate colormaps."""
import math

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


def plot_colormap(cmap, dpi=200):
    fig, ax = plt.subplots(figsize=(4, 0.4), dpi=dpi)
    mpl.colorbar.ColorbarBase(ax, cmap=cmap, orientation="horizontal")
    ax.set_title(cmap.name, fontsize=10, weight="bold")
    plt.show()


def plot_colormaps(cmaps, cols=None, subplot_size=None, dpi=200):
    # Define subplot_size
    if subplot_size is None:
        subplot_size = (2, 0.5)

    # Define number of subplots
    n = len(cmaps)

    # Define a layout most similar to a square
    if cols is None:
        cols = math.ceil(math.sqrt(n))
        cols = min(cols, 6)

    # Define number of rows required
    rows = int(np.ceil(n / cols))

    # Define figure width and height
    fig_width = cols * subplot_size[0]
    fig_height = rows * subplot_size[1]

    # Create dummy image for colormap
    im = np.outer(np.ones(10), np.arange(100))

    # Initialize figure
    fig, axes = plt.subplots(rows, cols, figsize=(fig_width, fig_height), dpi=dpi)
    fig.subplots_adjust(left=0, right=1, bottom=0, top=1, hspace=0.1, wspace=0.1)

    # Flatten axes for easy iteration
    axes = axes.ravel()

    # Loop through colormaps and axes
    for cmap, ax in zip(cmaps, axes):
        ax.set_title(cmap.name, fontsize=10, weight="bold")
        ax.imshow(im, cmap=cmap)
        ax.axis("off")  # Set axis off

    # Turn off any remaining axes
    for ax in axes[n:]:
        ax.axis("off")

    plt.show()


def show_colormap(cmap):
    from pycolorbar import get_cmap

    cmap = get_cmap(cmap)
    plot_colormap(cmap)


def show_colormaps(category=None, include_reversed=False, cols=None, subplot_size=None):
    from pycolorbar import available_colormaps, get_cmap

    # Retrieve colormaps to display
    names = available_colormaps(category=category, include_reversed=include_reversed)
    cmaps = [get_cmap(name) for name in sorted(names)]

    # Display colormaps
    if len(cmaps) > 0:
        plot_colormaps(cmaps, cols=cols, subplot_size=subplot_size)
    else:
        print(f"No colormaps are available within the category '{category}'.")



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to visualize univariate colorbars."""
import math

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


def _draw_colorbar(plot_kwargs, cbar_kwargs, fig, ax=None, cax=None):
    # Retrieve ticklabels (not accepted yet by fig.colorbar)
    ticklabels = cbar_kwargs.pop("ticklabels", None)
    # Display colorbar
    cbar = fig.colorbar(
        mpl.cm.ScalarMappable(**plot_kwargs),
        ax=ax,
        cax=cax,
        orientation="horizontal",
        **cbar_kwargs,
    )
    if ticklabels is not None:
        cbar.set_ticklabels(ticklabels)
    return cbar


def plot_colorbar(plot_kwargs, cbar_kwargs, ax=None, subplot_size=(6, 1)):
    # Initialize figure if necessary
    if ax is None:
        fig, ax = plt.subplots(figsize=subplot_size, layout="constrained")
    # Draw figure
    _ = _draw_colorbar(plot_kwargs=plot_kwargs, cbar_kwargs=cbar_kwargs, fig=fig, ax=None, cax=ax)
    plt.show()


def plot_colorbars(list_args, cols=None, subplot_size=None, dpi=200):
    # Define subplot_size
    if subplot_size is None:
        subplot_size = (5, 1.2)  # 3 --> 2

    # Define number of subplots
    n = len(list_args)

    # Define a layout most similar to a square
    if cols is None:
        cols = math.ceil(math.sqrt(n))
        cols = min(cols, 2)

    # Define number of rows required
    rows = int(np.ceil(n / cols))

    # Define figure width and height
    fig_width = cols * subplot_size[0]
    fig_height = rows * subplot_size[1]

    # Initialize figure
    fig, axes = plt.subplots(rows, cols, figsize=(fig_width, fig_height), dpi=dpi)

    # Flatten axes for easy iteration
    axes = axes.ravel()

    # Loop through colorbars and axes
    for idx, ((name, plot_kwargs, cbar_kwargs), ax) in enumerate(zip(list_args, axes)):
        _ = _draw_colorbar(plot_kwargs=plot_kwargs, cbar_kwargs=cbar_kwargs, fig=fig, ax=None, cax=ax)
        ax.set_title(name, fontsize=10, weight="bold")
    # Turn off any remaining axes
    for ax in axes[n:]:
        ax.axis("off")

    fig.tight_layout()
    plt.show()


def show_colorbar(name=None, user_plot_kwargs={}, user_cbar_kwargs={}, fig_size=(6, 1)):
    from pycolorbar import colorbars

    colorbars.show_colorbar(
        name=name,
        user_plot_kwargs=user_plot_kwargs,
        user_cbar_kwargs=user_cbar_kwargs,
        fig_size=fig_size,
    )


def show_colorbars(category=None, exclude_referenced=True, subplot_size=None):
    from pycolorbar import colorbars

    colorbars.show_colorbars(category=category, exclude_referenced=exclude_referenced, subplot_size=subplot_size)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
from importlib.metadata import PackageNotFoundError, version

from pycolorbar.settings.colorbar_registry import (  # noqa
    ColorbarRegistry,
    available_colorbars,
    get_cbar_dict,
    get_plot_kwargs,
    register_colorbar,
    register_colorbars,
)
from pycolorbar.settings.colorbar_validator import validate_cbar_dict  # noqa
from pycolorbar.settings.colorbar_visualization import show_colorbar, show_colorbars  # noqa
from pycolorbar.settings.colormap_registry import (  # noqa
    ColormapRegistry,
    available_colormaps,
    get_cmap,
    get_cmap_dict,
    register_colormap,
    register_colormaps,
)
from pycolorbar.settings.colormap_validator import validate_cmap_dict  # noqa
from pycolorbar.settings.colormap_visualization import show_colormap, show_colormaps  # noqa

# Create a module-level instance of ColormapRegistry
colormaps = ColormapRegistry.get_instance()

# Create a module-level instance of ColorbarRegistry
colorbars = ColorbarRegistry.get_instance()

__all__ = []

# Get version
try:
    __version__ = version("pycolorbar")
except PackageNotFoundError:
    # package is not installed
    pass



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
import numpy as np


class ColorEncoderDecoder:
    def __init__(self, external_data_range, internal_data_range, name):
        """
        Initialize a color encoder-decoder to convert color values between external and internal representations.

        Parameters
        ----------
        external_data_range : dict
            Dictionary specifying the data range for each channel in the external representation.
        internal_data_range : dict
            Dictionary specifying the data range for each channel in the internal representation.
        name : str
            The name of the color space.

        Raises
        ------
        ValueError
            If the keys of external_data_range and internal_data_range do not match.
        """
        if external_data_range.keys() != internal_data_range.keys():
            raise ValueError("The keys of external and internal data ranges must match.")

        self.external_data_range = external_data_range
        self.internal_data_range = internal_data_range
        self.decoding_functions = {channel: self._default_decode for channel in self.external_data_range}
        self.encoding_functions = {channel: self._default_encode for channel in self.internal_data_range}
        self.ndim = len(external_data_range)
        self.name = name

    def decode(self, colors):
        """
        Decode color values from external to internal representation.

        Parameters
        ----------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.

        Returns
        -------
        np.ndarray
            Decoded color values in internal representation.
        """
        return np.array(
            [
                self.decoding_functions[channel](
                    val, *self.external_data_range[channel], *self.internal_data_range[channel]
                )
                for channel, val in zip(self.external_data_range.keys(), colors.T)
            ]
        ).T

    def encode(self, colors):
        """
        Encode color values from internal to external representation.

        Parameters
        ----------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.

        Returns
        -------
        np.ndarray
            Encoded color values in external representation.
        """
        encoded_values = np.array(
            [
                self.encoding_functions[channel](
                    val, *self.internal_data_range[channel], *self.external_data_range[channel]
                )
                for channel, val in zip(self.internal_data_range.keys(), colors.T)
            ]
        ).T
        return encoded_values

    def _default_decode(self, value, from_min, from_max, to_min, to_max):
        """Default decoding function (linear scaling)."""
        return ((value - from_min) / (from_max - from_min)) * (to_max - to_min) + to_min

    def _default_encode(self, value, from_min, from_max, to_min, to_max):
        """Default encoding function (linear scaling)."""
        return ((value - from_min) / (from_max - from_min)) * (to_max - to_min) + to_min

    def check_colors(self, colors):
        """
        Check colors array dimension, size and type validity.

        Parameters
        ----------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.

        Returns
        -------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.

        """
        if not isinstance(colors, np.ndarray):
            raise ValueError("The colors array must be a numpy 2-dimensional array.")

        # Check 2D array
        if colors.ndim != 2:
            raise ValueError("The colors array must be a 2-dimensional array.")

        # Check number of columns
        if colors.shape[1] != self.ndim:
            raise ValueError(f"The colors array in the {self.name} color space must have {self.ndim} columns.")

        # Check data type
        expected_type = (np.int_, np.float_)
        if not issubclass(colors.dtype.type, expected_type):
            str_type = str(expected_type)
            raise ValueError(f"The colors array values must be of type {str_type}.")
        return colors

    def check_valid_internal_data_range(self, colors):
        """
        Check if the color values are within the internal data range for each channel.
        Raises an informative ValueError if a channel does not comply with the data range.

        Parameters
        ----------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.

        Raises
        ------
        ValueError
            If any channel values are not within the internal data range.
        """
        colors = self.check_colors(colors)
        for idx, channel in enumerate(self.internal_data_range.keys()):
            channel_colors = colors[:, idx]
            min_val, max_val = self.internal_data_range[channel]
            if not ((min_val <= channel_colors) & (channel_colors <= max_val)).all():
                raise ValueError(
                    f"Channel '{channel}' values are not within the internal data range. "
                    f"Expected range ({min_val}, {max_val}), but got values outside this range."
                )

    def check_valid_external_data_range(self, colors, strict=False):
        """
        Check if the color values are within the external data range for each channel.
        Raises an informative ValueError if a channel does not comply with the data range.
        If 'strict' is True, it ensures that not all values are within the internal data range.

        Parameters
        ----------
        colors : np.ndarray
            2D array where each column represents a channel in the color space.
        strict : bool, optional
            If True, performs a strict check to ensure that the values are within the external data range
            and not within the internal data range. Default is False.

        Raises
        ------
        ValueError
            If any channel values are not within the external data range or, if 'strict' is True,
            that not all channel values are also within the internal data range.
        """
        colors = self.check_colors(colors)
        for idx, channel in enumerate(self.external_data_range.keys()):
            channel_colors = colors[:, idx]
            min_val, max_val = self.external_data_range[channel]
            if not ((min_val <= channel_colors) & (channel_colors <= max_val)).all():
                raise ValueError(
                    f"Channel '{channel}' values are not within the external data range. "
                    f"Expected range ({min_val}, {max_val}), but got values outside this range."
                )

            if strict:
                internal_min_val, internal_max_val = self.internal_data_range[channel]
                if ((internal_min_val <= channel_colors) & (channel_colors <= internal_max_val)).all():
                    raise ValueError(
                        f"All '{channel}' values are within the internal data range "
                        "while expecting external representation."
                    )

    def is_within_internal_data_range(self, colors):
        """
        Check if the color values are within the internal data range for each channel.

        Parameters
        ----------
        colors : np.array
            2D-Array with the columns representing the color space dimensions.

        Returns
        -------
        bool
            True if all channel values are within the internal data range, False otherwise.
        """
        colors = self.check_colors(colors)
        conditions = [
            np.logical_and(
                colors[:, idx] >= self.internal_data_range[channel][0],
                colors[:, idx] <= self.internal_data_range[channel][1],
            )
            for idx, channel in enumerate(self.internal_data_range.keys())
        ]
        return np.all(np.vstack(conditions))

    def is_within_external_data_range(self, colors, strict: bool = False):
        """
        Check if the color values of each channels are within the external data range.
        Optionally, perform a strict check to ensure that not all values are also within the internal data range.

        Parameters
        ----------
        colors : np.array
            2D array where each column represents a channel in the color space.
        strict : bool, optional
            If True, performs a check to ensure that not all values are also within the internal data range.
            Default is False.

        Returns
        -------
        bool
            If strict=False, True if all channel values are within the external data range, False otherwise.
            If strict=True,  True if all channels values are within the external data range and
            not also all inside the internal data range, False otherwise.
        """
        colors = self.check_colors(colors)
        conditions = [
            np.logical_and(
                colors[:, idx] >= self.external_data_range[channel][0],
                colors[:, idx] <= self.external_data_range[channel][1],
            )
            for idx, channel in enumerate(self.external_data_range.keys())
        ]
        is_within = np.all(np.vstack(conditions))
        if not strict:
            return is_within
        else:
            return is_within and not self.is_within_internal_data_range(colors)


class RGBEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding RGB color values.

    This class handles the conversion between external and internal RGB color values.

    External data range: R (0-255), G (0-255), B (0-255)
    Internal data ranges: R (0-1), G (0-1), B (0-1)
    """

    def __init__(self):
        external_data_range = {"R": (0, 255), "G": (0, 255), "B": (0, 255)}
        internal_data_range = {"R": (0, 1), "G": (0, 1), "B": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="RGB")


class RGBAEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding RGBA color values.

    This class handles the conversion between external and internal RGB color values.

    External data range: R (0-255), G (0-255), B (0-255), A (0-100)
    Internal data ranges: R (0-1), G (0-1), B (0-1), A (0-1)
    """

    def __init__(self):
        external_data_range = {"R": (0, 255), "G": (0, 255), "B": (0, 255), "A": (0, 100)}
        internal_data_range = {"R": (0, 1), "G": (0, 1), "B": (0, 1), "A": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="RGBA")


class HSVEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding HSV (also called HSB) color values.

    This class handles the conversion between external and internal HSV color values.

    External data range: Hue (0-360), Saturation (0-100),  Value (0-100)
    Internal data range: Hue (0-2π),  Saturation (0-1),  Value (0-1)

    The Hue channel requires special handling for the conversion between degrees and radians.
    """

    def __init__(self):
        external_data_range = {"H": (0, 360), "S": (0, 100), "V": (0, 100)}
        internal_data_range = {"H": (0, 2 * np.pi), "S": (0, 1), "V": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="HSV")

        # Override default functions for the 'H' channel
        self.decoding_functions["H"] = self._hue_decode
        self.encoding_functions["H"] = self._hue_encode

    def _hue_decode(self, hue, from_min, from_max, to_min, to_max):
        """Custom decode function for Hue channel (from degrees to radians)."""
        return hue * (2 * np.pi) / 360  # Convert degrees to radians

    def _hue_encode(self, hue, from_min, from_max, to_min, to_max):
        """Custom encode function for Hue channel (from radians to degrees)."""
        return hue * 360 / (2 * np.pi)  # Convert radians to degrees


class HCLEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding HCL color values.

    The color space is also known as PolarLUV or cylindrical transformations of the CIELUV (CIELChuv)

    HCL rearrange the two CIELUV U and V chroma values into chroma (C) and hue (h).
    The CIELUV coordinate L* (luminance) remains unchanged.

    This class handles the conversion between external and internal HCL color values.

    External data range: Hue (0-255), Chroma (0-255), Luminance (0-255)
    Internal data range: Hue (0-1), Chroma (0-1), Luminance (0-1)

    """

    # TODO: check data_range and name
    def __init__(self):
        external_data_range = {"H": (0, 255), "C": (0, 255), "L": (0, 255)}
        internal_data_range = {"H": (0, 1), "C": (0, 1), "L": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="HCL")


class LCHEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding LCH color values.

    The color space is also known as PolarLAB or cylindrical transformations of the CIELAB (CIELChab)

    LCH rearrange the two CIELAB A and B chroma values into chroma (C) and hue (h).
    The CIELAB coordinate L* (luminance) remains unchanged.

    This class handles the conversion between external and internal LCh color values.

    External data range: Luminance (0-100), Chroma (0-200),  Hue (0-360)
    Internal data range: Luminance (0-1), Chroma (0-2),  Hue (0-2π)

    The Hue channel requires special handling for the conversion between degrees and radians.
    """

    # TODO: check data_range and name
    def __init__(self):
        external_data_range = {"L": (0, 100), "C": (0, 200), "H": (0, 360)}
        internal_data_range = {"L": (0, 1), "C": (0, 1), "H": (0, 2 * np.pi)}
        super().__init__(external_data_range, internal_data_range, name="LCH")

        # Override default functions for the 'H' channel
        self.decoding_functions["H"] = self._hue_decode
        self.encoding_functions["H"] = self._hue_encode

    def _hue_decode(self, hue, from_min, from_max, to_min, to_max):
        """Custom decode function for Hue channel (from degrees to radians)."""
        return hue * (2 * np.pi) / 360  # Convert degrees to radians

    def _hue_encode(self, hue, from_min, from_max, to_min, to_max):
        """Custom encode function for Hue channel (from radians to degrees)."""
        return hue * 360 / (2 * np.pi)  # Convert radians to degrees


class CIELUVEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding CIE LUV color values.

    This class handles the conversion between external and internal LUV color values.

    External data range: Luminance (0-100), U (-100,100),  V (-100, 100)
    Internal data range: Luminance (0-1),  U (-1,1),  V (-1,1)

    U* and V* can theoretically range from -100 to +100 but practical ranges are smaller.
    The u and v coordinates measure positions on green/red and blue/yellow axes.

    """

    def __init__(self):
        external_data_range = {"L": (0, 100), "U": (-100, 100), "V": (-100, 100)}
        internal_data_range = {"L": (0, 1), "U": (-1, 1), "V": (-1, 1)}
        super().__init__(external_data_range, internal_data_range, name="CIELUV")


class CIELABEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding CIE LAB color values.

    This class handles the conversion between external and internal CIELAB color values.

    External data range: Luminance (0-100), A (-128,127),  B (-128, 127)
    Internal data range: Luminance (0-1),  A (-1,1),  B (-1,1)

    Note: A and B can theoretically range from -128 to +127 but practical ranges are smaller.
    """

    def __init__(self):
        external_data_range = {"L": (0, 100), "A": (-128, 127), "B": (-128, 127)}
        internal_data_range = {"L": (0, 1), "A": (-1, 1), "B": (-1, 1)}  # TBD
        super().__init__(external_data_range, internal_data_range, name="CIELAB")


class CIEXYZEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding CIE XYZ color values.

    This class handles the conversion between external and internal CIE XYZ color values.

    External data range: X (0-100), Y (0-100),  Z (0-100)
    Internal data range: X (0-1), Y (0-1),  Z (0-1)

    """

    def __init__(self):
        external_data_range = {"X": (0, 100), "Y": (0, 100), "Z": (0, 100)}
        internal_data_range = {"X": (0, 1), "Y": (0, 1), "Z": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="CIEXYZ")


class CMYKEncoderDecoder(ColorEncoderDecoder):
    """
    A class for encoding and decoding CMYK color values.

    This class handles the conversion between external and internal CMYK color values.

    External data range: Cyan (0-100), Magenta (0-100), Yellow (0-100) Key/Black (0-100)
    Internal data range: Cyan (0-1), Magenta (0-1), Yellow (0-1) Key/Black (0-1)

    """

    def __init__(self):
        external_data_range = {"C": (0, 100), "M": (0, 100), "Y": (0, 100), "K": (0, 100)}
        internal_data_range = {"C": (0, 1), "M": (0, 1), "Y": (0, 1), "K": (0, 1)}
        super().__init__(external_data_range, internal_data_range, name="CMYK")


def _get_color_space_dict():
    class_dict = {
        "RGB": RGBEncoderDecoder,
        "RGBA": RGBAEncoderDecoder,
        "HSV": HSVEncoderDecoder,
        "LCH": LCHEncoderDecoder,
        "HCL": HCLEncoderDecoder,
        "CIELUV": CIELUVEncoderDecoder,
        "CIELAB": CIELABEncoderDecoder,
        "CIEXYZ": CIEXYZEncoderDecoder,
        "CMYK": CMYKEncoderDecoder,
    }
    return class_dict


def get_color_space_class(color_space):
    """
    Retrieve the class associated with the specified color space.

    Parameters
    ----------
    color_space : str
        The name of the color space.
        Valid color spaces are 'RGB', 'RGBA', 'HSV', 'LCH', 'HCL', 'CIELUV', 'CIELAB', 'CIEXYZ', 'CMYK'.

    Returns
    -------
    class
        The class corresponding to the specified color space.
    """
    class_dict = _get_color_space_dict()
    if color_space not in class_dict:
        raise ValueError(f"Color space '{color_space}' is not recognized.")
    return class_dict[color_space]


COLOR_SPACES = list(_get_color_space_dict())


def decode_colors(colors, color_space):
    """
    Decode colors from external to internal representation for the specified color space.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.

    Returns
    -------
    np.ndarray
        Decoded color values in internal representation.
    """
    color_space = color_space.upper()
    if color_space in COLOR_SPACES:
        color_class = get_color_space_class(color_space)
        return color_class().decode(colors)
    return colors


def encode_colors(colors, color_space):
    """
    Encode colors from internal to external representation for the specified color space.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.

    Returns
    -------
    np.ndarray
        Encoded color values in external representation.
    """
    color_space = color_space.upper()
    if color_space in COLOR_SPACES:
        color_class = get_color_space_class(color_space)
        return color_class().encode(colors)
    return colors


def is_within_internal_data_range(colors, color_space):
    """
    Check if the color values are within the internal data range for the specified color space.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.

    Returns
    -------
    bool
        True if all channel values are within the internal data range, False otherwise.
    """
    color_instance = get_color_space_class(color_space)()
    return color_instance.is_within_internal_data_range(colors)


def is_within_external_data_range(colors, color_space, strict=False):
    """
    Check if the color values are within the external data range for the specified color space.
    Optionally, perform a strict check to ensure that not all values are also within the internal data range.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.
    strict : bool, optional
        If True, performs a check to ensure that not all values are also within the internal data range.
        Default is False.

    Returns
    -------
    bool
        True if all channel values are within the external data range (and, if strict
        is True, not within the internal data range), False otherwise.
    """
    color_instance = get_color_space_class(color_space)()
    return color_instance.is_within_external_data_range(colors, strict=strict)


def check_valid_internal_data_range(colors, color_space):
    """
    Check if the color values are within the internal data range for the specified color space.
    Raises an informative ValueError if a channel does not comply with the data range.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.

    Raises
    ------
    ValueError
        If any channel values are not within the internal data range.
    """
    color_instance = get_color_space_class(color_space)()
    color_instance.check_valid_internal_data_range(colors)


def check_valid_external_data_range(colors, color_space, strict=False):
    """
    Check if the color values are within the external data range for each channel.
    Raises an informative ValueError if a channel does not comply with the data range.
    If 'strict' is True, it ensures that not all values are within the internal data range.

    Parameters
    ----------
    colors : np.ndarray
        2D array where each column represents a channel in the color space.
    color_space : str
        The name of the color space.
    strict : bool, optional
        If True, performs a strict check to ensure that the values are within the external data range
        and not within the internal data range. Default is False.

    Raises
    ------
    ValueError
        If any channel values are not within the external data range or, if 'strict' is True,
        that not all channel values are also within the internal data range.
    """
    color_instance = get_color_space_class(color_space)()
    color_instance.check_valid_external_data_range(colors, strict=strict)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Implementation of pydantic validator for univariate colormap YAML files."""
import re
from typing import Optional, Union

import numpy as np
from pydantic import BaseModel, field_validator

from pycolorbar.colors.colors_io import check_valid_external_data_range, check_valid_internal_data_range
from pycolorbar.utils.mpl import get_mpl_named_colors


def get_valid_color_space():
    """Get list of valid color spaces."""
    valid_names = [
        "hex",
        "name",
        "rgb",
        "rgba",
        "hcl",
        "lch",
        "hsv",
        "cieluv",
        "cielab",
        "ciexyz",
        "cmyk",
    ]
    return valid_names


def check_color_space(color_space):
    """Check color space validity."""
    valid_names = get_valid_color_space()
    if color_space.lower() not in valid_names:
        raise ValueError(f"Invalid color_space '{color_space}'. The supported color spaces are {valid_names}.")
    return color_space.lower()


def is_monotonically_increasing(values):
    return all(x <= y for x, y in zip(values, values[1:]))


class ColorMapValidator(BaseModel):
    """
    A validator for colormap configurations using Pydantic.

    Validates the fields of a colormap configuration, including the type of colormap,
    the color space, the colors themselves, and the alpha transparency settings.

    Attributes
    ----------
    type : str
        The type of the colormap (e.g., "ListedColormap", "LinearSegmentedColormap").
    color_space : str
        The color space of the colormap (e.g., "rgb", "hsv").
    colors : np.ndarray
        The array of colors defined for the colormap.
    colors_decoded: bool
        If True, assumes that the colors have been already decoded (internal representation).
        If False, assumes that the colors have not been decoded (external representation).
        The default is True.

    Methods
    -------
    validate_type(cls, v):
        Validates the type field.
    validate_color_space(cls, v):
        Validates the color_space field.
    validate_colors(cls, v, values):
        Validates the colors field.
    validate_segmentdata(cls, v, values)
        Validates the segmentdata field.
    """

    # NOTE: The order here governs the call and validation order of below methods

    # Internal flag
    colors_decoded: Optional[bool] = True

    # Mandatory colormap fields
    type: str
    color_space: str

    # LinearSegmentedColormap options
    segmentdata: Optional[dict] = None
    # gamma: Optional[float] = None

    # Optional colormap fields
    colors: Optional[Union[np.ndarray, list]] = None  # mandatory if segmentdata not provided !
    n: Optional[float] = None
    auxiliary: Optional[dict] = {}  # auxiliary information of the colormap (not checked !)

    # --------------------------------------------------
    # TODO:
    # - rename color_space as colors_space or colors_type ?
    # - interpolation_space? 'rgb', ...
    # - add gamma option
    # - correctly support segmentdata

    class Config:
        arbitrary_types_allowed = True

    @field_validator("type")
    def validate_type(cls, v):
        valid_types = [
            "ListedColormap",
            "LinearSegmentedColormap",
        ]
        assert isinstance(v, str), "Colormap 'type' must be a string."
        assert v in valid_types, f"Colormap 'type' must be one of {valid_types}"
        return v

    @field_validator("color_space")
    def validate_color_space(cls, v):
        check_color_space(color_space=v)
        return v

    @field_validator("colors_decoded")
    def validate_colors_decoded(cls, v):
        assert isinstance(v, bool), "colors_decoded must be a boolean."
        return v

    @field_validator("colors")
    def validate_colors(cls, v, values):
        if v is not None:
            v = np.asanyarray(v)
            color_space = values.data.get("color_space", "")
            assert len(v) > 0, "The 'colors' array must not be empty."
            assert len(v) > 1, "The 'colors' array must have at least 2 colors."
            validate_colors_values(v, color_space=color_space, decoded_colors=values.data.get("colors_decoded"))
        if v is None:
            assert (
                values.data.get("segmentdata", None) is not None
            ), "'colors' must be provided if 'segmentdata' is not specified"
        return v

    @field_validator("segmentdata")
    def validate_segmentdata(cls, v, values):
        if v is not None:
            assert (
                values.data.get("type") == "LinearSegmentedColormap"
            ), "'segmentdata' requires the 'type' 'LinearSegmentedColormap'."

            # Check the keys
            # - alpha can also be provided
            required_keys = ["red", "green", "blue"]
            if any(key not in v for key in required_keys):
                raise ValueError(f"'segmentdata' dictionary must contain keys: {required_keys}.")

            # Validate structure and monotonically increasing positions
            for key in required_keys:
                if not all(isinstance(item, tuple) and len(item) == 3 for item in v[key]):
                    raise ValueError(f"Each item in '{key}' must be a tuple of three floats.")
                positions = [item[0] for item in v[key]]
                if not is_monotonically_increasing(positions):
                    raise ValueError(f"Positions in '{key}' must be monotonically increasing.")

            # TODO:
            # - Check colors validity (encodded/decoded)
            # - Support encoding/decoding of the dictionary
            # - Allow for all color spaces !
            # - Currently support only rgb and rgba !

        return v


def validate_cmap_dict(cmap_dict: dict, decoded_colors=True):
    """
    Check the validity of a colormap dictionary.

    Parameters
    ----------
    cmap_dict : dict
        Colormap dictionary.
    decoded_colors : bool, optional
        Whether the colors are decoded (internal representation) or not. The default is True.

    Returns
    -------
    cmap_dict : TYPE
        Validated colormap dictionary.

    """
    # TODO: currently assumes that colors are already decoded (i.e. in 0-1 range for RGB)]
    # TODO: set defaults with pydantic?
    # --> Return what ColorMapValidator returns?

    # Set flag for color validation
    cmap_dict["colors_decoded"] = decoded_colors

    # Validate dictionary
    cmap_dict = ColorMapValidator(**cmap_dict).dict()

    # Remove flag for color validation
    _ = cmap_dict.pop("colors_decoded")

    # Return dictionary
    return cmap_dict


####-------------------------------------------------------------------------------------------------------------------.


def _check_ndim(colors: np.ndarray, expected_ndim: int):
    """
    Checks if the colors array has the expected number of dimensions.

    Parameters
    ----------
    colors : np.ndarray
        The array of colors to validate.
    expected_ndim : int
        The expected number of dimensions for the colors array.

    Raises
    ------
    ValueError
        If the colors array does not have the expected number of dimensions.
    """
    if colors.ndim != expected_ndim:
        raise ValueError(f"Colors array must be {expected_ndim}-D.")


def _check_type(colors: np.ndarray, expected_type: type):
    """
    Checks if the color values in the colors array are of the expected type.

    Parameters
    ----------
    colors : np.ndarray
        The array of colors to validate.
    expected_type : type
        The expected data type(s) of the color values.

    Raises
    ------
    ValueError
        If the color values are not of the expected type.
    """
    if not issubclass(colors.dtype.type, expected_type):
        str_type = str(expected_type)
        raise ValueError(f"Color values must be of type {str_type}.")


def validate_hex_colors(colors: np.ndarray) -> bool:
    """
    Validates the array of HEX colors.

    Parameters
    ----------
    colors : np.ndarray
        The array of colors to validate.

    Raises
    ------
    ValueError
        If the colors array is not 1-D or if any color is not a valid hex string.
    """

    _check_ndim(colors, 1)
    _check_type(colors, np.str_)
    hex_color_pattern = re.compile(r"^#(?:[0-9a-fA-F]{3}){1,2}$")
    if not all(hex_color_pattern.match(color) for color in colors):
        raise ValueError(
            "Invalid color format for 'hex'. "
            "Colors should be strings starting with '#' and followed by 3 or 6 hex digits."
        )


def validate_name_colors(colors: np.ndarray) -> bool:
    """
    Validates the array of named colors.

    For more info on named colors, see https://matplotlib.org/stable/gallery/color/named_colors.html

    Parameters
    ----------
    colors : np.ndarray
        The array of colors to validate.

    Raises
    ------
    ValueError
        If the colors array is not 1-D, if color values are not strings,
        or if any color name is not a valid named color.
    """
    _check_ndim(colors, 1)
    _check_type(colors, np.str_)
    valid_named_colors = get_mpl_named_colors()
    invalid_colors = colors[~np.isin(colors, valid_named_colors)]
    if len(invalid_colors) > 0:
        raise ValueError(f"Invalid named colors: {invalid_colors}.")


def validate_colors_values(colors, color_space, decoded_colors=True):
    """
    Validates the colors array based on the specified color space.

    Parameters
    ----------
    colors : np.ndarray
        The array of colors to validate.
    color_space : str
        The color space of the colors array (e.g., "hex", "rgb", "rgba", etc.).
    decoded_colors: bool
        If True, assumes that the colors are decoded (internal representation).
        If False, assumes that the colors are not decoded (external representation).

    Raises
    ------
    ValueError
        If the color_space is not supported or if the colors array fails
        validation checks for the specified color space.
    """
    # Check valid color space
    color_space = check_color_space(color_space)
    # Check valid color values
    if color_space == "name":
        validate_name_colors(colors)
    elif color_space == "hex":
        validate_hex_colors(colors)
    else:
        if decoded_colors:
            check_valid_internal_data_range(colors, color_space=color_space.upper())
        else:
            check_valid_external_data_range(colors, color_space=color_space.upper())


####-------------------------------------------------------------------------------------------------------------------.



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to read and write colormap YAML files."""
import os

import numpy as np

from pycolorbar.colors.colors_io import decode_colors as decode_colors_array
from pycolorbar.colors.colors_io import encode_colors as encode_colors_array
from pycolorbar.settings.colormap_validator import validate_cmap_dict
from pycolorbar.utils.yaml import read_yaml, write_yaml


def _get_colors_and_space(cmap_dict):
    if "colors" not in cmap_dict:
        raise KeyError("The colormap dictionary does not contain the 'colors' key.")
    if "color_space" not in cmap_dict:
        raise KeyError("The colormap dictionary does not contain the 'color_space' key.")
    colors = cmap_dict["colors"]
    color_space = cmap_dict["color_space"]
    return colors, color_space


def decode_colors(cmap_dict):
    colors, color_space = _get_colors_and_space(cmap_dict)
    # Convert colors to numpy array
    colors = np.asanyarray(colors)
    # Rescale colors to expected data range (i.e. rgb: 0-255 --> 0-1)
    colors = decode_colors_array(colors, color_space=color_space)
    # Return cmap_dict
    cmap_dict["colors"] = colors
    return cmap_dict


def encode_colors(cmap_dict):
    colors, color_space = _get_colors_and_space(cmap_dict)
    # Encode colors to expected YAML data range (i.e. rgb: 0-1 --> 0-255)
    colors = encode_colors_array(colors, color_space=color_space)
    # Reformat np.array to list of list
    if isinstance(colors, np.ndarray):
        if colors.ndim in [1, 2]:
            colors = colors.tolist()
            cmap_dict["colors"] = colors
        else:
            raise ValueError("Invalid 'colors' numpy array. Should be either 1D or 2D.")
    return cmap_dict


def read_cmap_dict(filepath):
    cmap_dict = read_yaml(filepath)
    cmap_dict = decode_colors(cmap_dict)
    cmap_dict = validate_cmap_dict(cmap_dict, decoded_colors=True)
    return cmap_dict


def write_cmap_dict(cmap_dict, filepath, force=False):
    # Check if file exist
    if os.path.exists(filepath):
        if force:
            os.remove(filepath)
        else:
            raise ValueError(f"The {filepath} already exists !")
    # Validate fields
    cmap_dict = validate_cmap_dict(cmap_dict=cmap_dict, decoded_colors=True)
    # Encode colors
    cmap_dict = encode_colors(cmap_dict)
    # Write file
    write_yaml(cmap_dict, filepath, sort_keys=False)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to retrieve the plotting arguments."""

import matplotlib as mpl
import numpy as np
from matplotlib.colors import (
    AsinhNorm,
    BoundaryNorm,
    CenteredNorm,
    LinearSegmentedColormap,
    LogNorm,
    NoNorm,
    Normalize,
    PowerNorm,
    SymLogNorm,
    TwoSlopeNorm,
)

import pycolorbar


def _get_cmap(cmap_settings, norm_settings):
    """Retrieve the colormap for a given colorbar setting."""

    name = cmap_settings.get("name")

    # Define number of colors required
    n = cmap_settings.get("n", None)
    if "labels" in cmap_settings:
        n = len(cmap_settings["labels"])
    if "boundaries" in norm_settings:
        n = len(norm_settings["boundaries"]) - 1

    # Define colormap
    if isinstance(name, str):
        cmap = pycolorbar.get_cmap(name=name, lut=n)
        return cmap
    else:  # Colormap combination
        # - Define default arguments
        if n is None:
            n = [None] * len(name)

        new_n = [256] * len(
            name
        )  # this could become a YAML parameter (divisible by len(names) (only for stacked cmaps))
        new_name = "_".join(name)
        # - Define combined colormap
        list_cmaps = [pycolorbar.get_cmap(cmap_name, lut=lut) for cmap_name, lut in zip(name, n)]
        cmap_colors = np.vstack([cmap(np.linspace(0, 1, lut)) for cmap, lut in zip(list_cmaps, new_n)])
        cmap = LinearSegmentedColormap.from_list(name=new_name, colors=cmap_colors)
        return cmap


def _finalize_cmap(cmap, cmap_settings):
    """Set alpha and under, over and bad colors."""
    # Set over and under colors
    # - If not specified, do not set ---> It will be filled with the first/last color value
    # - If 'none' --> It will be depicted in white
    if cmap_settings.get("over_color"):
        cmap.set_over(color=cmap_settings.get("over_color"), alpha=cmap_settings.get("over_alpha"))
    if cmap_settings.get("under_color"):
        cmap.set_under(color=cmap_settings.get("under_color"), alpha=cmap_settings.get("under_alpha"))

    # Set (bad) color for masked values
    # - If alpha not 0, can cause cartopy bug ?
    # --> https://stackoverflow.com/questions/60324497/specify-non-transparent-color-for-missing-data-in-cartopy-map
    if cmap_settings.get("bad_color"):
        cmap.set_bad(
            color=cmap_settings.get("bad_color"),
            alpha=cmap_settings.get("bad_alpha"),
        )
    return cmap


def get_cmap(cbar_dict):
    """Retrieve the colormap from a colorbar configuration dictionary."""
    # Retrieve settings
    cmap_settings = cbar_dict["cmap"]
    norm_settings = cbar_dict.get("norm", {})
    # Retrieve cmap
    cmap = _get_cmap(cmap_settings=cmap_settings, norm_settings=norm_settings)
    ### Set bad, under and over colors and transparency
    cmap = _finalize_cmap(cmap, cmap_settings)
    return cmap


####-------------------------------------------------------------------------------------------.
#### CategoryNorm
def create_category_norm(labels, first_value=0):
    """Define a BoundaryNorm that deal with categorical data."""
    n_labels = len(labels)
    norm_bins = np.arange(first_value - 0.01, n_labels + first_value) + 0.5
    norm = mpl.colors.BoundaryNorm(boundaries=norm_bins, ncolors=n_labels)
    return norm


####-------------------------------------------------------------------------------------------.
#### Norm utility


def get_norm_function(name):
    """Retrieve the norm function."""
    # Define norm function mapping
    norm_functions = {
        "Norm": Normalize,
        "NoNorm": NoNorm,
        "BoundaryNorm": BoundaryNorm,
        "TwoSlopeNorm": TwoSlopeNorm,
        "CenteredNorm": CenteredNorm,
        "LogNorm": LogNorm,
        "SymLogNorm": SymLogNorm,
        "PowerNorm": PowerNorm,
        "AsinhNorm": AsinhNorm,
        "CategoryNorm": create_category_norm,
    }
    return norm_functions[name]


def get_norm(cbar_dict):
    """Define the norm instance."""
    norm_settings = cbar_dict["norm"]
    norm_name = norm_settings.get("name", "Norm")
    # Retrieve norm function
    norm_func = get_norm_function(norm_name)
    # Retrieve norm arguments
    norm_kwargs = norm_settings.copy()
    _ = norm_kwargs.pop("name", None)
    # Set default values for BoundaryNorm
    if norm_name == "BoundaryNorm" and norm_kwargs.get("ncolors", None) is None:
        boundaries = norm_settings["boundaries"]
        ncolors = len(boundaries) - 1
        norm_kwargs["ncolors"] = ncolors
    # Define norm
    norm = norm_func(**norm_kwargs)
    return norm


####-------------------------------------------------------------------------------------------.
#### pycolorbar default settings


# TODO:
# Check vmin and vmax are None if using BoundaryNorm
# Check vmin and vmax are None when providing a Norm

# --> Allow labels also for BoundaryNorm ? But remove from kwargs when norm generation ?

# --> Adapt cmap for 'labels' (define n)


def get_plot_cbar_kwargs(cbar_dict):
    # ------------------------------------------------------------------------.
    # Set default colormap
    if cbar_dict == {}:
        cbar_dict["cmap"] = {"name": "jet"}
        cbar_dict["norm"] = {"name": "Norm"}

    # ------------------------------------------------------------------------.
    # Initialize kwargs
    plot_kwargs = {}
    cbar_kwargs = get_default_cbar_kwargs()

    # ------------------------------------------------------------------------.
    # Define cmap, norm, ticks and cbar appearance based on colorbar dictionary settings
    if "cmap" in cbar_dict:
        plot_kwargs["cmap"] = get_cmap(cbar_dict)
    if "norm" in cbar_dict:
        norm = get_norm(cbar_dict)
        ticks, ticklabels = _get_ticks_settings(cbar_dict, norm=norm)
        plot_kwargs["norm"] = norm
        # - TODO: Add only if not None ?
        cbar_kwargs["ticks"] = ticks
        cbar_kwargs["ticklabels"] = ticklabels
    if "cbar" in cbar_dict:
        cbar_kwargs.update(cbar_dict["cbar"])

    # ------------------------------------------------------------------------.
    return plot_kwargs, cbar_kwargs


def get_default_cbar_kwargs():
    cbar_kwargs = {
        "ticks": None,
        "ticklocation": "auto",
        "spacing": "uniform",  # or proportional
        "extend": "neither",
        "extendfrac": "auto",
        "extendrect": False,
        "label": None,
        "drawedges": False,
        "shrink": 1,
    }
    return cbar_kwargs


def _dynamic_formatting_floats(float_array):
    """Function to format the floats defining the class limits of the colorbar."""
    float_array = np.array(float_array, dtype=float)
    labels = []
    for label in float_array:
        if 0.1 <= label < 1:
            formatting = ",.2f"
        elif 0.01 <= label < 0.1:
            formatting = ",.2f"
        elif 0.001 <= label < 0.01:
            formatting = ",.3f"
        elif 0.0001 <= label < 0.001:
            formatting = ",.4f"
        elif label >= 1 and label.is_integer():
            formatting = "i"
        else:
            formatting = ",.1f"
        if formatting != "i":
            labels.append(format(label, formatting))
        else:
            labels.append(str(int(label)))
    return labels


def _get_ticks_settings(cbar_dict, norm=None):
    # Retrieve settings
    norm_settings = cbar_dict["norm"]
    boundaries = norm_settings.get("boundaries", None)
    labels = norm_settings.get("labels", None)
    # Define ticks and ticklabels for BoundaryNorm instances
    if isinstance(norm, BoundaryNorm):
        # Define ticks
        if boundaries is not None:  # BoundaryNorm
            ticks = boundaries
        else:  # CategoryNorm
            ticks = norm.boundaries[:-1] + 0.5
        # Define ticklabels
        if labels is None:  # BoundaryNorm
            # Generate color level strings with correct amount of decimal places
            ticklabels = _dynamic_formatting_floats(ticks)  # [f"{tick:.1f}" for tick in ticks] # for 0.1 probability
        else:
            # Define tick formatter and ticks
            fmt = mpl.ticker.FuncFormatter(lambda x, pos: labels[norm(x)])
            ticklabels = [fmt(tick) for tick in ticks]
    else:
        ticks = None
        ticklabels = None
    return ticks, ticklabels


####--------------------------------------------------------------------------------------------.
#### Update pycolorbar settings based on user arguments


def update_plot_cbar_kwargs(default_plot_kwargs, default_cbar_kwargs, user_plot_kwargs={}, user_cbar_kwargs={}):
    # If no user kwargs, return default kwargs
    if user_plot_kwargs == {} and user_cbar_kwargs == {}:
        return default_plot_kwargs, default_cbar_kwargs

    # If user cmap
    # - is a string, retrieve colormap
    # - is None --> delete the argument
    _parse_user_cmap(user_plot_kwargs=user_plot_kwargs)

    # If norm is specified, vmin and vmax must be None !
    _check_no_vmin_vmax_if_norm_specified(user_plot_kwargs=user_plot_kwargs)

    # TODO
    # - WHAT TO DO IF TICKS AND TICKLABELS SPECIFIED IN user_cbar_kwargs
    # --> If BoundaryNorm and change length --> Raise error?
    # --> If no BoundaryNorm?

    # Deal with categorical/discrete colorbar (when user specify new norm)
    # - Remove ticks and ticklabels when user specify new norm !
    # - Later on:
    #   - If user specify a new cmap --> the cmap is resampled based on len(ticklabels)
    #   - If vmin or vmax are specified --> a Normalize(vmin, vmax) replace BoundaryNorm
    if user_plot_kwargs.get("norm", None) is not None:
        _remove_defaults_ticks_and_ticklabels(default_cbar_kwargs=default_cbar_kwargs)

    # Deal with categorical/discrete labeled colorbar (when user provides a new cmap)
    # - Resample user-provided colormap
    if default_cbar_kwargs.get("ticklabels", None) is not None:
        _resample_user_cmap_for_labeled_colorbar(
            user_plot_kwargs=user_plot_kwargs, default_cbar_kwargs=default_cbar_kwargs
        )

    # Deal with xarray user_plot_kwargs 'levels' option
    # - Define a BoundaryNorm and resample the cmap accordingly
    if "levels" in user_plot_kwargs:
        _create_boundary_norm_from_levels(user_plot_kwargs=user_plot_kwargs, default_plot_kwargs=default_plot_kwargs)

    # If norm is not specified but vmin and vmax are specified
    # --> Check the default norm accepts vmin and vmax arguments
    # --> If yes, update the default norm to use specified vmin and vmax
    # --> If no, warn and define a Normalize(vmin, vmax)
    if user_plot_kwargs.get("norm", None) is None:
        _update_default_norm_using_vmin_and_vmax(
            user_plot_kwargs=user_plot_kwargs, default_plot_kwargs=default_plot_kwargs
        )

    # Drop vmin and vmax from user_plot_kwargs (not accepted by PolyCollection)
    _ = user_plot_kwargs.pop("vmin", None)
    _ = user_plot_kwargs.pop("vmax", None)

    # Deal with xarray optional 'extend' plot_kwargs
    # - extend is copied also in the user_cbar_kwargs
    if "extend" in user_plot_kwargs and "extend" not in user_cbar_kwargs:
        user_cbar_kwargs["extend"] = user_plot_kwargs["extend"]

    # Update defaults with custom kwargs
    default_plot_kwargs.update(user_plot_kwargs)
    default_cbar_kwargs.update(user_cbar_kwargs)

    return default_plot_kwargs, default_cbar_kwargs


def _remove_defaults_ticks_and_ticklabels(default_cbar_kwargs):
    default_ticks = default_cbar_kwargs.get("ticks", None)
    default_ticklabels = default_cbar_kwargs.get("ticklabels", None)
    if default_ticks is not None or default_ticklabels is not None:
        default_cbar_kwargs.pop("ticks", None)
        default_cbar_kwargs.pop("ticklabels", None)


def _resample_user_cmap_for_labeled_colorbar(user_plot_kwargs, default_cbar_kwargs):
    default_ticklabels = default_cbar_kwargs.get("ticklabels", None)
    n_labels = len(default_ticklabels)
    if user_plot_kwargs.get("cmap", None) is not None:
        cmap = user_plot_kwargs["cmap"]
        if n_labels != cmap.N:
            cmap = cmap.resampled(n_labels)
            user_plot_kwargs["cmap"] = cmap


def _create_boundary_norm_from_levels(user_plot_kwargs, default_plot_kwargs):
    # xarray user_plot_kwargs 'levels' option
    # - Resample cmap and define BoundaryNorm

    # Get user settings
    vmin = user_plot_kwargs.get("vmin", None)
    vmax = user_plot_kwargs.get("vmax", None)
    levels = user_plot_kwargs["levels"]

    # Check norm is not defined
    if "norm" in user_plot_kwargs:
        raise ValueError("Either specify 'norm' or 'levels'.")
    # Define boundaries
    if isinstance(levels, int):
        if vmin is None or vmax is None:
            raise ValueError("If 'levels' is an integer, you must specify 'vmin' and 'vmax'.")
        boundaries = np.linspace(vmin, vmax, levels)
    else:
        boundaries = list(levels)
    ncolors = len(boundaries) - 1
    # Define boundary norm
    norm = BoundaryNorm(boundaries=boundaries, ncolors=ncolors)
    # Resample colormap
    if "cmap" in user_plot_kwargs:
        user_plot_kwargs["cmap"] = user_plot_kwargs["cmap"].resampled(ncolors)
    else:
        default_plot_kwargs["cmap"] = default_plot_kwargs["cmap"].resampled(ncolors)
    # Add "BoundaryNorm" to user_plot_kwargs
    user_plot_kwargs["norm"] = norm
    # Remove "levels"
    _ = user_plot_kwargs.pop("levels")


def _update_default_norm_using_vmin_and_vmax(user_plot_kwargs, default_plot_kwargs):
    vmin = user_plot_kwargs.get("vmin", None)
    vmax = user_plot_kwargs.get("vmax", None)

    if vmin is not None or vmax is not None:
        # If default accept vmin, vmax --> update vmin/vmax attributes
        try:
            if vmin is not None:
                default_plot_kwargs["norm"].vmin = vmin
            if vmax is not None:
                default_plot_kwargs["norm"].vmax = vmax
            user_plot_kwargs["norm"] = default_plot_kwargs["norm"]
        # Otherwise define a new Normalize norm
        except Exception:
            norm_class = type(default_plot_kwargs["norm"])
            print(
                f"The default pycolorbar norm is a {norm_class} and does not accept 'vmin' and 'vmax'.\n "
                f"Switching the norm to Normalize(vmin={vmin}, vmax={vmax}) !"
            )
            user_plot_kwargs["norm"] = Normalize(vmin=vmin, vmax=vmax)


def _check_no_vmin_vmax_if_norm_specified(user_plot_kwargs):
    vmin = user_plot_kwargs.get("vmin", None)
    vmax = user_plot_kwargs.get("vmax", None)
    norm = user_plot_kwargs.get("norm", None)
    if norm is not None:
        if vmin is not None or vmax is not None:
            raise ValueError("If the 'norm' is specified, 'vmin' and 'vmax' must not be specified.")


def _parse_user_cmap(user_plot_kwargs):
    cmap = user_plot_kwargs.get("cmap", None)
    if isinstance(cmap, str):
        user_plot_kwargs["cmap"] = pycolorbar.get_cmap(name=cmap)
    if cmap is None:
        _ = user_plot_kwargs.pop("cmap", None)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np


def get_mpl_named_colors():
    """
    Retrieves the list of valid named colors available in matplotlib.

    Returns
    -------
    np.ndarray
        An array of valid named colors.

    Notes
    -----
    This function combines colors from TABLEAU_COLORS, BASE_COLORS, CSS4_COLORS, and XKCD_COLORS in matplotlib.
    """
    named_colors = list(mcolors.get_named_colors_mapping())
    # named_colors = (
    #     list(mcolors.TABLEAU_COLORS) +
    #     list(mcolors.BASE_COLORS)    +
    #     list(mcolors.CSS4_COLORS)    +
    #     list(mcolors.XKCD_COLORS)
    # )
    named_colors = np.array(named_colors, dtype=str)
    return named_colors


def get_mpl_colormaps():
    return plt.colormaps()



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Implementation of pydantic validator for univariate colorbar YAML files."""

import re
from typing import Optional, Union

import numpy as np
from pydantic import BaseModel, field_validator, model_validator

from pycolorbar.utils.mpl import get_mpl_colormaps, get_mpl_named_colors

####---------------------------------------------------------------------------------------------------------.
#### Colormap Settings


class ColormapSettings(BaseModel):
    name: Union[str, list[str]]
    n: Optional[Union[int, list[int]]] = None
    bad_color: Optional[Union[str, list, tuple]] = None
    bad_alpha: Optional[float] = None
    over_color: Optional[Union[str, list, tuple]] = None
    over_alpha: Optional[float] = None
    under_color: Optional[Union[str, list, tuple]] = None
    under_alpha: Optional[float] = None

    class Config:
        arbitrary_types_allowed = True

    @field_validator("name")
    def validate_name(cls, v):
        """Check if cmap is a registered matplotlib colormap or name in pycolorbar.colormaps.registry."""
        import pycolorbar

        if isinstance(v, str):
            valid_names = get_mpl_colormaps() + pycolorbar.colormaps.names
            assert v in valid_names, f"'{v}' is not a recognized colormap name."
        elif isinstance(v, list):
            for name in v:
                valid_names = get_mpl_colormaps() + pycolorbar.colormaps.names
                assert name in valid_names, f"'{name}' is not a recognized colormap name."
        return v

    @field_validator("n")
    def validate_n(cls, v, values, **kwargs):
        if v is not None:
            validated_settings = values.data
            # Single colormap
            if isinstance(validated_settings.get("name"), str):
                assert isinstance(v, int) and v > 0, "'n' must be a positive integer."
            # Multiple colormaps
            if isinstance(validated_settings.get("name"), list):
                assert len(validated_settings.get("name")) == len(
                    v
                ), "'n' must match the number of color maps in 'name'."
                for n in v:
                    assert isinstance(n, int) and n > 0, "'n' values must be positive integers."
        return v

    @field_validator("bad_color", "over_color", "under_color")
    def validate_colors(cls, v):
        if v is not None:
            if isinstance(v, str):
                if v == "none":
                    return v
                # Check if it's a named color
                if v in get_mpl_named_colors():
                    return v
                # Check if it's a hex color
                hex_color_pattern = re.compile(r"^#(?:[0-9a-fA-F]{3}){1,2}$")
                if not hex_color_pattern.match(v):
                    raise ValueError(
                        'Invalid color format. Expected hex string like "#RRGGBB" or "#RRGGBBAA", or a named color.'
                    )
            elif isinstance(v, (list, tuple)) and len(v) in [3, 4]:
                # Check if it's an RGB or RGBA tuple
                if not all(
                    isinstance(color_component, (int, float)) and 0 <= color_component <= 1 for color_component in v
                ):
                    raise ValueError("Invalid RGB/RGBA format. Expected tuple with values between 0 and 1.")
            else:
                raise ValueError("Invalid color format. Expected a named color, hex string, or RGB/RGBA tuple.")
        return v

    @field_validator("bad_alpha", "under_alpha", "over_alpha")
    def validate_bad_alpha(cls, v):
        if v is not None:
            assert 0 <= v <= 1, "bad_alpha must be between 0 and 1"
        return v


####-------------------------------------------------------------------------------------------------------------------.
#### Norm Settings


def _check_norm_invalid_args(norm_name, args, valid_args):
    invalid_keys = set(args) - set(valid_args)
    if invalid_keys:
        raise ValueError(f"Invalid parameters {invalid_keys} for normalization type '{norm_name}'.")


def _check_vmin_vcenter_vmax(vmin, vcenter, vmax, norm_name):
    if vmin is not None and vcenter is not None:
        assert vmin < vcenter, "'vmin' must be less than 'vcenter' for 'TwoSlopeNorm'."
    if vmax is not None and vcenter is not None:
        assert vcenter < vmax, "'vmax' must be larger than 'vcenter' 'TwoSlopeNorm'."


def _check_vmin_vmax(vmin, vmax):
    assert isinstance(vmin, (int, float, type(None))), "'vmin' must be an integer, float or None."
    assert isinstance(vmax, (int, float, type(None))), "'vmax' must be an integer, float or None."
    if vmin is not None and vmax is not None:
        assert vmin < vmax, "vmin must be less than vmax."


def _check_clip(clip):
    assert isinstance(clip, bool), "'clip' must be either True or False."


def _check_extend(extend):
    valid_extends = ["neither", "both", "min", "max"]
    assert extend in valid_extends, f"Invalid extend option '{extend}'. Valid options are {valid_extends}."


def _is_monotonically_increasing(x):
    x = np.asanyarray(x)
    return np.all(x[1:] > x[:-1])


def _get_boundary_norm_expected_ncolors(norm_settings):
    boundaries = norm_settings.get("boundaries", [])
    extend = norm_settings.get("extend", "neither")
    if extend == "neither":
        required_ncolors = len(boundaries) - 1
    elif extend in ["min", "max"]:
        required_ncolors = len(boundaries)
    else:  #  extend == 'both':
        required_ncolors = len(boundaries) + 1
    return required_ncolors


class NormalizeSettings(BaseModel):
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for Normalize."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for Normalize."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in Normalize."""
        valid_args = {"vmin", "vmax", "clip"}
        _check_norm_invalid_args(norm_name="Normalize", args=values.keys(), valid_args=valid_args)
        return values


class CategoryNormSettings(BaseModel):
    labels: list[str]
    first_value: Optional[int] = 0

    @field_validator("labels")
    def validate_labels(cls, v, values):
        """Validate labels for CategoryNorm."""
        if v is not None:
            assert isinstance(v, list), "'labels' must be a list."
            assert len(v) >= 2, "'labels' must have at least two strings"
            assert all([isinstance(label, str) for label in v]), "'labels' must be a list of strings."
        return v

    @field_validator("first_value")
    def validate_first_value(cls, v, values):
        """Validate first_value for CategoryNorm."""
        if v is not None:
            assert isinstance(v, int), "'first_value' must be an integer."
        return v

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in Normalize."""
        valid_args = {"labels", "first_value"}
        _check_norm_invalid_args(norm_name="CategoryNorm", args=values.keys(), valid_args=valid_args)
        return values


class BoundaryNormSettings(BaseModel):
    # "ncolors" if not specified is taken from len(boundaries)
    boundaries: list[float]
    clip: Optional[bool] = False
    extend: Optional[str] = "neither"
    ncolors: Optional[int] = None

    @field_validator("boundaries")
    def validate_boundaries(cls, v):
        """Validate `boundaries` list for BoundaryNorm."""
        assert isinstance(v, list), "'boundaries' list is required for 'BoundaryNorm'."
        assert all(isinstance(b, (int, float)) for b in v), "'boundaries' must be a list of numbers."
        assert _is_monotonically_increasing(v), "'boundaries' must be monotonically increasing."
        return v

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate ` clip` option for BoundaryNorm."""
        _check_clip(v)
        return v

    @field_validator("extend")
    def validate_extend(cls, v):
        """Validate `extend` option for BoundaryNorm."""
        if v is not None:
            _check_extend(v)
        return v

    @field_validator("ncolors")
    def validate_ncolors(cls, v, values):
        """Validate `ncolors` for BoundaryNorm."""
        if v is not None:
            assert isinstance(v, int), "'ncolors' must be an integer for 'BoundaryNorm'."
            assert v >= 2, "'ncolors' must be equal or larger than 2."
            # - If extend is "neither" (default) there must be equal or larger than len(boundaries) - 1 colors.
            # - If extend is "min" or "max" ncolors must be equal or larger than len(boundaries)
            # - If extend is "both"  ncolors must be equal or larger than len(boundaries) + 1
            validated_settings = values.data
            extend = validated_settings.get("extend", "neither")
            required_ncolors = _get_boundary_norm_expected_ncolors(norm_settings=validated_settings)
            if extend == "neither":
                assert (
                    v >= required_ncolors
                ), f"'ncolors' must be equal or larger than len('boundaries') - 1 ({required_ncolors})."
            elif extend in ["min", "max"]:
                assert (
                    v >= required_ncolors
                ), f"'ncolors' must be equal or larger than len('boundaries') ({required_ncolors})."
            elif extend == "both":
                assert (
                    v >= required_ncolors
                ), f"'ncolors' must be equal or larger than len('boundaries') + 1 ({required_ncolors})."
        return v

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in BoundaryNorm."""
        valid_args = {"boundaries", "ncolors", "clip", "extend"}
        _check_norm_invalid_args(norm_name="BoundaryNorm", args=values.keys(), valid_args=valid_args)
        return values


class NoNormSettings(BaseModel):
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for NoNorm."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for NoNorm."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in NoNorm."""
        valid_args = {"vmin", "vmax", "clip"}
        _check_norm_invalid_args(norm_name="NoNorm", args=values.keys(), valid_args=valid_args)
        return values


class CenteredNormSettings(BaseModel):
    vcenter: Optional[Union[int, float]] = 0
    halfrange: Optional[Union[int, float]] = None
    clip: Optional[bool] = False

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for CenteredNorm."""
        _check_clip(v)
        return v

    @field_validator("vcenter")
    def validate_vcenter(cls, v):
        """Validate `vcenter` for CenteredNorm."""
        assert isinstance(v, (int, float)), "'vcenter' must be an integer or float."
        return v

    @field_validator("halfrange")
    def validate_halfrange(cls, v):
        """Validate `halfrange` for CenteredNorm."""
        if v is not None:
            assert isinstance(v, (int, float)), "'halfrange' must be an integer, float or None."
        return v

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in CenteredNorm."""
        valid_args = {"vcenter", "halfrange", "clip"}
        _check_norm_invalid_args(norm_name="CenteredNorm", args=values.keys(), valid_args=valid_args)
        return values


class TwoSlopeNormSettings(BaseModel):
    vcenter: float
    vmin: Optional[float] = None
    vmax: Optional[float] = None

    @field_validator("vcenter")
    def validate_vcenter(cls, v):
        """Validate `vcenter` for TwoSlopeNorm."""
        assert isinstance(v, (int, float)), "'vcenter' must be an integer or float."
        return v

    @model_validator(mode="before")
    def check_vmin_vcenter_vmax(cls, values):
        """Check `vmin`, `vcenter`, and `vmax` for TwoSlopeNorm."""
        vmin, vcenter, vmax = values.get("vmin"), values.get("vcenter"), values.get("vmax")
        _check_vmin_vcenter_vmax(vmin=vmin, vcenter=vcenter, vmax=vmax, norm_name="TwoSlopeNorm")
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in TwoSlopeNorm."""
        valid_args = {"vcenter", "vmin", "vmax"}
        _check_norm_invalid_args(norm_name="TwoSlopeNorm", args=values.keys(), valid_args=valid_args)
        return values


class LogNormSettings(BaseModel):
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for LogNorm."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for LogNorm."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        if vmin is not None:
            if vmin <= 0:
                raise ValueError("LogNorm vmin should be a positive value.")
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in LogNorm."""
        valid_args = {"vmin", "vmax", "clip"}
        _check_norm_invalid_args(norm_name="LogNorm", args=values.keys(), valid_args=valid_args)
        return values


class SymLogNormSettings(BaseModel):
    linthresh: float
    linscale: Optional[float] = 1.0
    base: Optional[float] = 10
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("linthresh")
    def validate_linthresh(cls, v):
        """Validate `linthresh` for SymLogNorm."""
        assert v > 0, "'linthresh' must be positive for 'SymLogNorm'."
        return v

    @field_validator("linscale", "base")
    def validate_linscale_base(cls, v, field):
        """Validate `linscale` and `base` for SymLogNorm."""
        if v is not None:
            assert v > 0, f"'{field.name}' must be positive for 'SymLogNorm'."
        return v

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for SymLogNorm."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for SymLogNorm."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in SymLogNorm."""
        valid_args = ["linthresh", "linscale", "vmin", "vmax", "clip", "base"]
        _check_norm_invalid_args(norm_name="SymLogNorm", args=values.keys(), valid_args=valid_args)
        return values


class PowerNormSettings(BaseModel):
    gamma: float
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("gamma")
    def validate_gamma(cls, v):
        """Validate `gamma` for PowerNorm."""
        assert isinstance(v, (int, float)), "'gamma' must be an integer or float."
        return v

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for PowerNorm."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for PowerNorm."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in PowerNorm."""
        valid_args = ["gamma", "vmin", "vmax", "clip"]
        _check_norm_invalid_args(norm_name="PowerNorm", args=values.keys(), valid_args=valid_args)
        return values


class AsinhNormSettings(BaseModel):
    linear_width: Optional[Union[int, float]] = 1
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    clip: Optional[bool] = False

    @field_validator("linear_width")
    def validate_linear_width(cls, v):
        """Validate `linear_width` for AsinhNorm."""
        assert isinstance(v, (int, float)), "'linear_width' must be an integer or float."
        return v

    @field_validator("clip")
    def validate_clip(cls, v):
        """Validate `clip` option for AsinhNorm."""
        _check_clip(v)
        return v

    @model_validator(mode="before")
    def check_vmin_vmax(cls, values):
        """Check `vmin` and `vmax` for AsinhNorm."""
        vmin, vmax = values.get("vmin"), values.get("vmax")
        _check_vmin_vmax(vmin, vmax)
        return values

    @model_validator(mode="before")
    def check_valid_args(cls, values):
        """Check for no excess parameters in AsinhNorm."""
        valid_args = ["linear_width", "vmin", "vmax", "clip"]
        _check_norm_invalid_args(norm_name="AsinhNorm", args=values.keys(), valid_args=valid_args)
        return values


def _check_valid_norm_name(name):
    valid_names = [
        "Norm",
        "NoNorm",
        "BoundaryNorm",
        "TwoSlopeNorm",
        "CenteredNorm",
        "LogNorm",
        "SymLogNorm",
        "PowerNorm",
        "AsinhNorm",
        "CategoryNorm",
    ]
    if name not in valid_names:
        raise ValueError(f"Invalid norm '{name}'. Valid options are {valid_names}.")


def check_norm_settings(norm_settings):
    # Check valid *Norm name
    norm_settings = norm_settings.copy()
    name = norm_settings.pop("name", "Norm")
    _check_valid_norm_name(name)
    # Define *Norm validators
    norm_settings_mapping = {
        "Norm": NormalizeSettings,
        "NoNorm": NoNormSettings,
        "BoundaryNorm": BoundaryNormSettings,
        "TwoSlopeNorm": TwoSlopeNormSettings,
        "CenteredNorm": CenteredNormSettings,
        "LogNorm": LogNormSettings,
        "SymLogNorm": SymLogNormSettings,
        "PowerNorm": PowerNormSettings,
        "AsinhNorm": AsinhNormSettings,
        "CategoryNorm": CategoryNormSettings,
    }
    # Retrieve NormSettings Validator
    validator = norm_settings_mapping[name]
    # Validate settings
    norm_settings = validator(**norm_settings).dict()
    # Return validated settings
    norm_settings["name"] = name
    return norm_settings


####-------------------------------------------------------------------------------------------------------------------.


class ColorbarSettings(BaseModel):
    extend: Optional[str] = "neither"
    extendfrac: Optional[Union[float, list[float], str]] = "auto"
    extendrect: Optional[bool] = False
    label: Optional[str] = None  # title of colorbar

    class Config:
        arbitrary_types_allowed = True

    @field_validator("extend")
    def validate_extend(cls, v):
        """Validate extend option."""
        if v is not None:
            _check_extend(v)
        return v

    @field_validator("extendfrac")
    def validate_extendfrac(cls, v):
        """Validate extend fraction."""
        if v is not None:
            if isinstance(v, list):
                assert all(
                    isinstance(frac, (float, int)) and 0 <= frac <= 1 for frac in v
                ), "Each extendfrac in the list must be a float or int between 0 and 1."
            elif isinstance(v, str):
                assert v == "auto", "'extendfrac' must not be a string."
            else:
                assert isinstance(v, (float, int)) and 0 <= v <= 1, "extendfrac must be a float or int between 0 and 1."
        return v

    @field_validator("extendrect")
    def validate_extendrect(cls, v):
        """Validate extend rectangle option."""
        if v is not None:
            assert isinstance(v, bool), "extendrect must be a boolean value."
        return v

    @field_validator("label")
    def validate_label(cls, v):
        """Validate label as string."""
        if v is not None:
            assert isinstance(v, str), "label must be a string."
        return v


####-------------------------------------------------------------------------------------------------------------------.


def resolve_colorbar_reference(cbar_dict, name, checked_references=None):
    import pycolorbar

    keys = list(cbar_dict)
    if len(keys) > 1:
        if np.any(np.isin(keys, ["auxiliary", "reference"], invert=True)):
            raise ValueError("If referencing another colorbar, only 'reference' and 'auxiliary' keys are allowed.")

    # Retrieve reference
    reference_name = cbar_dict["reference"]

    # Check reference is available
    if reference_name not in pycolorbar.colorbars.names:
        raise ValueError(f"The '{reference_name}' colorbar is not registered in pycolorbar. Invalid reference !")

    # Check for circular references
    if checked_references is None:
        checked_references = []
    if reference_name in checked_references:
        raise ValueError(f"Circular reference detected with '{reference_name}'.")

    # Retrieve new dictionary
    cbar_ref_dict = pycolorbar.colorbars.get_cbar_dict(reference_name, validate=False)

    # If another reference, visit recursively the references
    if "reference" in cbar_ref_dict:
        checked_references.append(name)
        return resolve_colorbar_reference(cbar_ref_dict, name=reference_name, checked_references=checked_references)

    # Return the original colorbar dictionary
    return cbar_ref_dict


def _check_discrete_norm_cmap_settings(cmap_settings, norm_settings):
    norm = norm_settings.get("name", "Norm")
    if norm not in ["CategoryNorm", "BoundaryNorm"]:
        return None

    n = cmap_settings.get("n", None)
    if n is not None:
        # Retrieve expected number of colors
        if norm == "CategoryNorm":
            ncolors = len(norm_settings["labels"])
        else:  # "BoundaryNorm"
            ncolors = _get_boundary_norm_expected_ncolors(norm_settings=norm_settings)
        # Check it match expectations
        # - Single Colormap
        if isinstance(n, int):
            assert n == ncolors, f"'n' is optional and must be {ncolors} for the specified discrete norm."
        # - Multiple colormaps
        else:
            assert sum(n) == ncolors, f"The sum of 'n' must be {ncolors} for the specified discrete norm."


def validate_cbar_dict(cbar_dict: dict, name: str, resolve_reference=False):
    """Validate a colorbar dictionary."""
    # Raise error for empty dictionary or wrong type
    if not isinstance(cbar_dict, dict):
        raise TypeError("The colorbar dictionary must be a dictionary.")
    if len(cbar_dict) == 0:
        raise ValueError("The colorbar dictionary can not be empty.")

    # Check if cbar_dict reference to another colorbar settings
    if "reference" in cbar_dict:
        referenced_cbar_dict = resolve_colorbar_reference(cbar_dict, name=name)
        if resolve_reference:
            cbar_dict = referenced_cbar_dict
        else:
            return cbar_dict

    # Retrieve cmap, norm and cbar settings
    cmap_settings = cbar_dict["cmap"]
    norm_settings = cbar_dict.get("norm", {})
    cbar_settings = cbar_dict.get("cbar", {})

    # Test validity
    invalid_configuration = False

    try:
        cmap_settings = ColormapSettings(**cmap_settings).dict()
    except Exception as e:
        invalid_configuration = True
        print(f"Colormap validation error: {e}")

    try:
        norm_settings = check_norm_settings(norm_settings)
    except Exception as e:
        invalid_configuration = True
        print(f"Norm validation error: {e}")

    try:
        cbar_settings = ColorbarSettings(**cbar_settings).dict()
    except Exception as e:
        invalid_configuration = True
        print(f"Colorbar validation error: {e}")

    # Consistency checks
    try:
        _check_discrete_norm_cmap_settings(cmap_settings=cmap_settings, norm_settings=norm_settings)
    except Exception as e:
        invalid_configuration = True
        print(f"Categorical Colormap validation error: {e}")

    if invalid_configuration:
        raise ValueError("Invalid configuration")

    # Return the validated dictionary
    cbar_dict["cmap"] = cmap_settings
    cbar_dict["norm"] = norm_settings
    cbar_dict["cbar"] = cbar_settings
    return cbar_dict



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to read and write colorbar YAML files."""
import os

import numpy as np

from pycolorbar.settings.colorbar_validator import validate_cbar_dict
from pycolorbar.utils.yaml import read_yaml, write_yaml


def remove_if_exists(filepath, force=False):
    if os.path.exists(filepath):
        if force:
            os.remove(filepath)
        else:
            raise ValueError(f"The {filepath} already exists !")


# CONFIG FILE. pycolorbar.validate_at_registration
# CONFIG FILE. pycolorbar.validate_at_selection


def read_cbar_dict(filepath, name=None):
    """Read colorbar YAML file with single colorbar settings."""
    cbar_dict = read_yaml(filepath)
    filename = os.path.basename(filepath)
    name = os.path.splitext(filename)[0]
    cbar_dict = validate_cbar_dict(cbar_dict, name=name)
    return cbar_dict


def is_single_colorbar_settings(dictionary):
    """Determine if a dictionary is a single colorbar settings."""
    if np.any(np.isin(["cmap", "norm", "cbar", "auxiliary"], list(dictionary))):
        return True
    else:
        return False


def read_cbar_dicts(filepath):
    """Read colorbar YAML file with a single or multiple colorbar settings."""
    dictionary = read_yaml(filepath)
    # If not single colorbar settings, returns the cbar_dicts
    if not is_single_colorbar_settings(dictionary):
        return dictionary
    # Otherwise retrieve colorbar name from filename
    filename = os.path.basename(filepath)
    name = os.path.splitext(filename)[0]
    # Return the setting in the cbar_dicts format
    return {name: dictionary}


def write_cbar_dict(cbar_dict, name, filepath, force=False):
    """Write a single colorbar settings dictionary to a YAML file.."""
    # Check if file exist
    remove_if_exists(filepath, force=force)
    # Validate fields
    cbar_dict = validate_cbar_dict(cbar_dict=cbar_dict, name=name)
    # Write file
    write_yaml(cbar_dict, filepath, sort_keys=False)


def tmp_conv_to_new_format(cbar_dict):
    # cmap
    cmap_keys = [
        "cmap",
        "cmap_n",
        "bad_alpha",
        "bad_color",
        "over_color",
        "under_color",
    ]

    cmap_keys_map = {
        "cmap": "name",
        "cmap_n": "n",
    }

    norm_keys = [
        "norm",
        "vmin",
        "vmax",
        "linthresh",
        "linscale",
        "base",
        "levels",
        "labels",
    ]

    norm_keys_map = {
        "norm": "name",
        "levels": "boundaries",
    }

    cbar_keys = [
        "extend",
        "extendfrac",
        "extendrect",
        "label",
    ]

    new_cbar_dict = {}

    # Reference
    if "reference" in cbar_dict:
        new_cbar_dict["reference"] = cbar_dict["reference"]
        return new_cbar_dict

    # Cmap
    if np.any(np.isin(cmap_keys, list(cbar_dict))):
        new_cbar_dict["cmap"] = {}
        for key in cmap_keys:
            if key in cbar_dict:
                value = cbar_dict[key]
                if key in cmap_keys_map:
                    new_key = cmap_keys_map[key]
                else:
                    new_key = key
                new_cbar_dict["cmap"][new_key] = value

    # Norm
    if np.any(np.isin(norm_keys, list(cbar_dict))):
        new_cbar_dict["norm"] = {}
        if "norm" not in cbar_dict:
            if "labels" in cbar_dict:
                new_cbar_dict["norm"]["name"] = "CategoryNorm"
            elif "levels" in cbar_dict:
                new_cbar_dict["norm"]["name"] = "BoundaryNorm"
            else:
                new_cbar_dict["norm"]["name"] = "Norm"

        for key in norm_keys:
            if key in cbar_dict:
                value = cbar_dict[key]
                if key in norm_keys_map:
                    new_key = norm_keys_map[key]
                else:
                    new_key = key
                new_cbar_dict["norm"][new_key] = value

    # Cbar
    if np.any(np.isin(cbar_keys, list(cbar_dict))):
        new_cbar_dict["cbar"] = {}
        for key in cbar_keys:
            if key in cbar_dict:
                value = cbar_dict[key]
                new_cbar_dict["cbar"][key] = value
    # Auxiliary
    aux_keys = ["citation", "citation_url", "author", "author_url", "comment", "category"]
    if np.any(np.isin(aux_keys, list(cbar_dict))):
        new_cbar_dict["auxiliary"] = {}
        for key in aux_keys:
            if key in cbar_dict:
                value = cbar_dict[key]
                new_cbar_dict["auxiliary"][key] = value

    return new_cbar_dict


def write_cbar_dicts(cbar_dicts, filepath, names=None, force=False, sort_keys=False):
    """Write a multiple colorbar settings dictionary to a YAML file."""
    if isinstance(names, str):
        names = [names]

    # Check if file exist
    remove_if_exists(filepath, force=force)

    # Select colorbars
    if names is not None:
        cbar_dicts = {name: cbar_dicts[name] for name in names}

    # Conversion colorbars
    # cbar_dicts = {name: tmp_conv_to_new_format(cbar_dict=cbar_dict) for name, cbar_dict in cbar_dicts.items()}

    # Validate colorbars
    cbar_dicts = {name: validate_cbar_dict(cbar_dict=cbar_dict, name=name) for name, cbar_dict in cbar_dicts.items()}

    # Write file
    write_yaml(cbar_dicts, filepath, sort_keys=sort_keys)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
"""Define functions to build colormaps in multiple color spaces."""

# import numpy as np
# import matplotlib as mpl
# import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap


def convert_colors(colors, color_space):
    #  TODO: IMPLEMENT !
    pass
    return colors


def create_cmap(cmap_dict, name):
    cmap_type = cmap_dict["type"]
    color_space = cmap_dict["color_space"]

    colors = cmap_dict.get("colors", None)
    segmentdata = cmap_dict.get("segmentdata", None)

    # Convert colors to interpolation space
    # - if ListedColormap --> RGBA
    # - if LinearSegmentedColormap --> interpolation_space (default RGBA)
    # --> TODO: or create ColorMap Classes interpolating in the <interpolation_space>
    colors = convert_colors(colors, color_space)

    # Create Colormap
    if cmap_type == "ListedColormap":
        n = cmap_dict.get("n", None)
        return ListedColormap(colors, name=name, N=n)
    else:
        n = cmap_dict.get("n", 256)
        gamma = cmap_dict.get("gamma", 1.0)
        if "segmentdata" not in cmap_dict:
            # Retrieve n colors in 'interpolation_space' (when type=LinearSegmentedColormap)
            # TODO

            # Retrieve colormap
            return LinearSegmentedColormap.from_list(name=name, colors=colors, N=n, gamma=gamma)
        else:
            segmentdata = cmap_dict["segmentdata"]
            return LinearSegmentedColormap(name=name, segmentdata=segmentdata, N=n, gamma=gamma)



# -----------------------------------------------------------------------------.
# MIT License

# Copyright (c) 2024 pycolorbar developers
#
# This file is part of pycolorbar.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# -----------------------------------------------------------------------------.
import glob
import os
import pathlib


def _recursive_glob(dir_path, glob_pattern):
    # ** search for all files recursively
    # glob_pattern = os.path.join(base_dir, "**", "metadata", f"{station_name}.yml")
    # metadata_filepaths = glob.glob(glob_pattern, recursive=True)

    dir_path = pathlib.Path(dir_path)
    return [str(path) for path in dir_path.rglob(glob_pattern)]


def list_paths(dir_path, glob_pattern, recursive=False):
    """Return a list of filepaths and directory paths."""
    if not recursive:
        return glob.glob(os.path.join(dir_path, glob_pattern))
    else:
        return _recursive_glob(dir_path, glob_pattern)


def list_files(dir_path, glob_pattern, recursive=False):
    """Return a list of filepaths (exclude directory paths)."""
    paths = list_paths(dir_path, glob_pattern, recursive=recursive)
    filepaths = [f for f in paths if os.path.isfile(f)]
    return filepaths
